/*
 * CertValue.java
 * PROJECT: JDigiDoc
 * DESCRIPTION: Digi Doc functions for creating
 *	and reading signed documents. 
 * AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
 *==================================================
 * Copyright (C) AS Sertifitseerimiskeskus
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * GNU Lesser General Public Licence is available at
 * http://www.gnu.org/copyleft/lesser.html
 *==================================================
 */
package ee.sk.digidoc;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;

import nu.xom.Attribute;
import nu.xom.Element;
import ee.sk.digidoc.utils.XMLObject;



/**
 * Models the ETSI <X509Certificate> 
 * and <EncapsulatedX509Certificate> elements.
 * Holds certificate data. Such elements will
 * be serialized under the <CertificateValues>
 * and <X509Data> elements
 * @version 2.0
 * @author SEMRYC
 */
public class CertValue extends XMLObject{
    /** elements id atribute if present */
    private String m_id;
    /** parent object - Signature ref */
    private Signature m_signature;
    /** CertID type - signer, responder, tsa */
    private int m_type;
    /** certificate */
    private X509Certificate m_cert;
    
    /** possible cert value type values */
    public static final int CERTVAL_TYPE_UNKNOWN = 0;
    public static final int CERTVAL_TYPE_SIGNER = 1;
    public static final int CERTVAL_TYPE_RESPONDER = 2;
    public static final int CERTVAL_TYPE_TSA = 3;

    /** 
     * Creates new CertValue 
     * and initializes everything to null
     */
    public CertValue() {
        m_id = null;
        m_signature = null;
        m_cert = null;
        m_type = CERTVAL_TYPE_UNKNOWN;
        try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
    }
    
    /**
     * Accessor for Signature attribute
     * @return value of Signature attribute
     */
    public Signature getSignature()
    {
    	return m_signature;
    }
    
    /**
     * Mutator for Signature attribute
     * @param uprops value of Signature attribute
     */
    public void setSignature(Signature sig)
    {
    	m_signature = sig;
    	try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
    }
    
    /**
     * Accessor for id attribute
     * @return value of certId attribute
     */
    public String getId() {
        return m_id;
    }
    
    /**
     * Mutator for id attribute
     * @param str new value for certId attribute
     */    
    public void setId(String str) 
    {
    	m_id = str;
    	try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
    }
    
    /**
     * Accessor for type attribute
     * @return value of type attribute
     */
    public int getType() {
        return m_type;
    }
    
    /**
     * Mutator for type attribute
     * @param n new value for issuer attribute
     * @throws DigiDocException for validation errors
     */    
    public void setType(int n) 
        throws DigiDocException
    {
        DigiDocException ex = validateType(n);
        if(ex != null)
            throw ex;
        m_type = n;
        try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
    }
    
    /**
     * Helper method to validate type
     * @param n input data
     * @return exception or null for ok
     */
    private DigiDocException validateType(int n)
    {
        DigiDocException ex = null;
        if(n < 0 || n > CERTVAL_TYPE_TSA)
            ex = new DigiDocException(DigiDocException.ERR_CERTID_TYPE, 
                "Invalid CertValue type", null);
        return ex;
    }    
        
    /**
     * Accessor for Cert attribute
     * @return value of Cert attribute
     */
    public X509Certificate getCert()
    {
    	return m_cert;
    }
    
    /**
     * Mutator for Cert attribute
     * @param uprops value of Cert attribute
     */
    public void setCert(X509Certificate cert)
    {
    	m_cert = cert;
    	try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
    }
    
    protected Element createElement() throws DigiDocException{
    	if(thisObject==null){
    		if(m_type == CERTVAL_TYPE_SIGNER) {
    			thisObject=makeEasyElement("X509Certificate",SignedDoc.XMLNS_XMLDSIG);
    		}else if(m_type == CERTVAL_TYPE_RESPONDER || m_type == CERTVAL_TYPE_TSA) {
    			thisObject=makeEasyElement("EncapsulatedX509Certificate",SignedDoc.XMLNS_XMLDSIG);
    			thisObject.addAttribute(new Attribute("Id",m_id));
    		}
    	}
    	if (m_cert!=null){
    		if (thisObject.getChildCount()>0)
    			thisObject.removeChildren();
    		byte[] encoded;
    		try {
    			encoded = m_cert.getEncoded();
    		} catch (CertificateEncodingException e) {
    			throw new DigiDocException(DigiDocException.ERR_CERT_UNKNOWN,"Certificate codification error.",e);
    		}
    		thisObject.appendChild(Base64Util.encode(encoded, encoded.length));
    	}
    	return thisObject;
    }
    
    protected Element createElement(String namespace) throws DigiDocException{
    	return createElement();
    }
    
}
