/* CLASS: Content (previously DataFile)
 * MODIFIED: It has been refactored in order to make a generic attached content
 * class. The tag name associated can be specified.
 * Copyright (C) 2007  Ministerio de Administraciones Publicas, Gobierno de Espaa
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details at 
 * http://www.gnu.org/licenses/lgpl.html
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package ee.sk.digidoc;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.util.ArrayList;

import nu.xom.Attribute;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.ParsingException;
import nu.xom.ValidityException;
import ee.sk.digidoc.factory.CanonicalizationFactory;
import ee.sk.digidoc.factory.XOMBuilderFactory;
import ee.sk.digidoc.utils.XMLObject;

/**
 * Represents the attached content (previously 
 * known as DataFile)
 * Only attached mode is possible at this version.
 * @version 1.0
 */

public class Content extends XMLObject implements Serializable {
	
	
	private static final long serialVersionUID = 1L;
	
	/** Id attribute od this DataFile */
	private String m_id;
	
	/** digest type of detatched file */
	private String m_digestType;
	
	/** digest value of detatched file */
	private byte[] m_digestValue;
	
	/** digest value of the XML form of <code><CONTENT></code>
	 * If read from XML file then calculated immediately
	 * otherwise on demand
	 */
	private byte[] m_origDigestValue;
	
	/** data file contents in original form */
	private byte[] m_body;
	
	//Solo util si tenemos XML que por ahora no. 
	//Pasamos siempre a UTF8, para encajar en nuestro XML root
	//private String m_codepage;
	
	/** parent object reference */
	private SignedDoc m_sigDoc;
	
	/** Content mode*/
	private String m_mode;
	
	/** the only allowed value for digest type */
	public static final String DIGEST_TYPE_SHA1 = "sha1";
	
	/**CONTENT Tag and attribute names*/
	private String tagName = "CONTENT";
	private String dataMimeType = null;
	private String hashAlgorithm = null; 
	private String m_encoding = null;
	
	/** 
	 * Creates new DataFile 
	 * @param id id of the DataFile
	 * @param contenType DataFile content type
	 * @param fileName original file name (without path!)
	 * @param mimeType contents mime type
	 * @param sdoc parent object
	 * @throws DigiDocException for validation errors
	 */
	public Content(String tag, String id, SignedDoc sdoc) 
	throws DigiDocException
	{
		m_id=id;
		m_sigDoc = sdoc;
		m_digestType = null;
		m_digestValue = null;
		m_body = null;
		m_origDigestValue = null;
		if (tag!=null)
			tagName=tag;
		m_mode=m_sigDoc.getSignatureMode();
		hashAlgorithm=null;
		createElement();
	}
	
	public Content(String tag, String id, SignedDoc doc, byte[] body)
	throws DigiDocException{
		tagName=tag;
		m_id=id;
		m_sigDoc=doc;
		m_body=body;
		m_mode=doc.getSignatureMode();
		hashAlgorithm=null;
		createElement();
	}
	
	public Content(String tag, String id, SignedDoc doc, byte[] body,String mimeType)
	throws DigiDocException{
		tagName=tag;
		m_id=id;
		m_sigDoc=doc;
		m_mode=doc.getSignatureMode();
		m_body=body;
		dataMimeType=mimeType;
		hashAlgorithm=null;
		createElement();
	}
	
	public Content(String tag, String id, SignedDoc doc, byte[] body,String mimeType, String algHash)
	throws DigiDocException{
		tagName=tag;
		m_id=id;
		m_sigDoc=doc;
		m_mode=doc.getSignatureMode();
		m_body=body;
		dataMimeType=mimeType;
		hashAlgorithm=algHash;
		createElement();
	}
	
	public void setDataMimeType(String mimeType){
		this.dataMimeType=mimeType;
		try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
	}
	
	public String getDataMimeType(){
		return dataMimeType;
	}
	
	/**
	 * Accessor for body attribute. 
	 * Note that the body is normally NOT LOADED
	 * from file and this attribute is empty!
	 * @return value of body attribute
	 */
	public byte[] getBody() {
		return m_body;
	}
	
	/**
	 *  Establece el contenido mediante una secuencia de bytes en BASE64
	 * @param data new value for body attribute
	 */    
	public void setBody(byte[] data) 
	{
		m_body = data;
		try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
	}
	
	/**
	 * Devuelve la cadena del contenido
	 * @return body as string
	 */
	public String getBodyAsString() 
	throws DigiDocException
	{
		String str;
		try {
			str = new String(m_body,"UTF8");
		} catch (UnsupportedEncodingException e) {
			throw new DigiDocException(DigiDocException.ERR_ENCODING,"Error encoding data body to UTF8",e);
		}
		return str;
	}
	
	/**
	 * Accessor for id attribute
	 * @return value of id attribute
	 */
	public String getId() {
		return m_id;
	}
	
	/**
	 * Mutator for id attribute
	 * @param str new value for id attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setId(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateId(str, false);
		if(ex != null)
			throw ex;
		m_id = str;
		createElement();
	}
	
	/**
	 * Helper method to validate an id
	 * @param str input data
	 * @param bStrong flag that specifies if Id atribute value is to
	 * be rigorously checked (according to digidoc format) or only
	 * as required by XML-DSIG
	 * @return exception or null for ok
	 */
	private DigiDocException validateId(String str, boolean bStrong)
	{
		DigiDocException ex = null;
		if(str == null)
			ex = new DigiDocException(DigiDocException.ERR_DATA_FILE_ID, 
					"Id is a required attribute", null);
		if(str != null && bStrong && 
				(str.charAt(0) != 'D' || !Character.isDigit(str.charAt(1))))
			ex = new DigiDocException(DigiDocException.ERR_DATA_FILE_ID, 
					"Id attribute value has to be in form D<number>", null);
		return ex;
	}
	
	/**
	 * Accessor for digestType attribute
	 * @return value of digestType attribute
	 */
	public String getDigestType() {
		return m_digestType;
	}
	
	/**
	 * Mutator for digestType attribute
	 * @param str new value for digestType attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setDigestType(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateDigestType(str);
		if(ex != null)
			throw ex;
		m_digestType = str;
	}
	
	/**
	 * Helper method to validate a digestType
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateDigestType(String str)
	{
		DigiDocException ex = null;
		if(str != null && !str.equals(DIGEST_TYPE_SHA1))
			ex = new DigiDocException(DigiDocException.ERR_DATA_FILE_DIGEST_TYPE, 
					"The only supported digest type is sha1", null);
		return ex;
	}
	
	/**
	 * Accessor for digestValue attribute
	 * @return value of digestValue attribute
	 */
	public byte[] getDigestValue() 
	throws DigiDocException
	{
		return m_digestValue;
	}
	
	/**
	 * Mutator for digestValue attribute
	 * @param data new value for digestValue attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setDigestValue(byte[] data) 
	throws DigiDocException
	{
		DigiDocException ex = validateDigestValue(data);
		if(ex != null)
			throw ex;
		m_digestValue = data;
	}
	
	/**
	 * Accessor for digest attribute
	 * @return value of digest attribute
	 */
	public byte[] getDigest() 
	throws DigiDocException
	{
		return calculateDigest();
	}
	
	/**
	 * Mutator for digest attribute
	 * @param data new value for digest attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setDigest(byte[] data) 
	throws DigiDocException
	{
		DigiDocException ex = validateDigestValue(data);
		if(ex != null)
			throw ex;
		m_origDigestValue = data;
	}
	
	/**
	 * Helper method to validate a digestValue
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateDigestValue(byte[] data)
	{
		DigiDocException ex = null;
		if(data != null && data.length != SignedDoc.SHA1_DIGEST_LENGTH)
			ex = new DigiDocException(DigiDocException.ERR_DATA_FILE_DIGEST_VALUE, 
					"SHA1 digest value must be 20 bytes", null);
		return ex;
	}
	
	
	
	/**
	 * Helper method to validate the whole
	 * DataFile object
	 * @param bStrong flag that specifies if Id atribute value is to
	 * be rigorously checked (according to digidoc format) or only
	 * as required by XML-DSIG
	 * @return a possibly empty list of DigiDocException objects
	 */
	public ArrayList validate(boolean bStrong)
	{
		ArrayList errs = new ArrayList();
		DigiDocException ex = validateId(m_id, bStrong);
		if(ex != null)
			errs.add(ex);
		ex = validateDigestType(m_digestType);
		if(ex != null)
			errs.add(ex);
		ex = validateDigestValue(m_digestValue);
		if(ex != null)
			errs.add(ex);
		return errs;
	}
	
	public byte[] calculateDigest()throws DigiDocException{
		if(m_origDigestValue == null) {
			byte[] tmp=canonicalizeXml(getElement());
			byte[] hash=SignedDoc.digest(tmp);
			return hash;
		}
		else
			return m_origDigestValue;
	}
	
	/**
	 * Helper method to canonicalize a piece of xml
	 * @param xml data to be canonicalized
	 * @return canonicalized xml
	 */
	private byte[] canonicalizeXml(Element data) {
		try {				 
			CanonicalizationFactory canFac = SignedDoc.getCanonicalizeInstance();
			byte[] tmp = canFac.canonicalize(data, SignedDoc.CANONICALIZATION_METHOD_20010315);
			return tmp;
		} catch(Exception ex) {
			System.out.println("Canonicalizing exception: " + ex);
		}
		return null;
	}

	protected Element createElement(String namespace) throws DigiDocException{
		return createElement();
	}
	
	private boolean containCharsBase64(byte[] input){
		//Esta funcion se ha creado poque se han encontrado cadenas que no estan en base64 y
		//sin embargo pueden decodificarse desde esta codificacion.
		String b64Chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwkyz123456789+/.=";
		
		String inStr = new String(input); 		
		for(int i=0; i<inStr.length(); i++){			
			if(b64Chars.indexOf(inStr.substring(i,i))==-1)
				return false;			
		}
		
		return true;
	}
	
	protected Element createElement() throws DigiDocException{
		try{
			if (m_mode.equals(SignedDoc.MODE_DETACHED_SIGNATURE)){
				if (thisObject==null)
					thisObject=new Element(tagName);
				if (m_id!=null&&m_body!=null){
					Builder builder=XOMBuilderFactory.getInstance(false);
					thisObject.addAttribute(new Attribute("Id",m_id));
					thisObject.removeChildren();
					try{
						Document tmp=builder.build(new ByteArrayInputStream(m_body));
						Element el=(Element)tmp.getChild(0);
						tmp.replaceChild(el,new Element("TMPRUSH"));
						thisObject.appendChild(el);
						dataMimeType="text/xml";
					}catch(Exception e1){
						try{//Si son datos en base64 los introducimos tal cual
							//Intentamos decodificar los datos en base64	
							if(containCharsBase64(m_body)){
								Base64Util.decode(m_body);
								thisObject.appendChild(new String(m_body));
							}
							else{//Si no son datos en base64 los codificamos a base64
								//Contiene caracteres que no son base64
								String b64Content=Base64Util.encode(m_body);
								thisObject.appendChild(b64Content);
							}
						}
						catch(Exception e2){//Si no son datos en base64 los codificamos a base64
							//No se ha podido decodificar. Procedemos a codificar en base64 el contenido
							String b64Content=Base64Util.encode(m_body);
							thisObject.appendChild(b64Content);
						}	
						if(m_encoding == null || !m_encoding.equals(SignedDoc.BASE64_ENCODING)){
							try{
								this.setDataEncoding(SignedDoc.BASE64_ENCODING);
							}
							catch(Exception e){
							}
						}
					}
				}
				if (dataMimeType!=null){
					thisObject.addAttribute(new Attribute("MimeType",dataMimeType));
				}else{
					thisObject.addAttribute(new Attribute("MimeType",SignedDoc.DEFAULT_MIMETYPE));
				}
				if (m_encoding!=null){
					thisObject.addAttribute(new Attribute("Encoding",m_encoding));
				}else {
					if (thisObject.getAttribute("Encoding")!=null)
						thisObject.removeAttribute(thisObject.getAttribute("Encoding"));
				}
				if (hashAlgorithm != null){
					thisObject.addAttribute(new Attribute("Algorithm",hashAlgorithm));
				}
			}else if (m_mode.equals(SignedDoc.MODE_ENVELOPED_SIGNATURE)){
				Builder builder=XOMBuilderFactory.getInstance(false);
				Document tmp=builder.build(new ByteArrayInputStream(m_body));
				thisObject=tmp.getRootElement();
				tmp.setRootElement(new Element("TMPRUSH"));
				thisObject.detach();
			}else if(m_mode.equals(SignedDoc.MODE_ENVELOPING_SIGNATURE)){
				if (thisObject==null)
					thisObject=makeEasyElement("Object",XMLNS_XMLDSIG);
				if (m_id!=null&&m_body!=null){
					Builder builder=XOMBuilderFactory.getInstance(false);
					thisObject.addAttribute(new Attribute("Id",m_id));
					thisObject.removeChildren();
					try{
						Document tmp=builder.build(new ByteArrayInputStream(m_body));
						Element el=(Element)tmp.getChild(0);
						tmp.replaceChild(el,new Element("TMPRUSH"));
						thisObject.appendChild(el);
						dataMimeType="text/xml";
					}catch(Exception e){
						String b64Content=Base64Util.encode(m_body);
						thisObject.appendChild(b64Content);
						if(m_encoding == null || !m_encoding.equals(SignedDoc.BASE64_ENCODING)){
							try{
								this.setDataEncoding(SignedDoc.BASE64_ENCODING);
							}
							catch(Exception e1){
							}
						}
					}
				}
				if (m_encoding!=null){
					thisObject.addAttribute(new Attribute("Encoding",m_encoding));
				}else {
					if (thisObject.getAttribute("Encoding")!=null)
						thisObject.removeAttribute(thisObject.getAttribute("Encoding"));
				}
				if (dataMimeType!=null){
					thisObject.addAttribute(new Attribute("MimeType",dataMimeType));
				}else{
					thisObject.addAttribute(new Attribute("MimeType",SignedDoc.DEFAULT_MIMETYPE));
				}
				if (hashAlgorithm != null){
					thisObject.addAttribute(new Attribute("Algorithm",hashAlgorithm));
				}
			}
		}catch(IOException e){
			throw new DigiDocException(DigiDocException.ERR_UNKNOWN,"IO error.",e);
		} catch (ValidityException e) {
			throw new DigiDocException(DigiDocException.ERR_XML_CONVERT,"Validation error.",e);
		} catch (ParsingException e) {
			throw new DigiDocException(DigiDocException.ERR_XML_CONVERT,"Parsing error: Invalid content type for selected signature mode",e);
		}
		return thisObject;
	}

	public String getDataEncoding() {
		return m_encoding;
	}

	public void setDataEncoding(String encoding) throws DigiDocException {
		this.m_encoding = encoding;
		createElement();
	}

	static Content parseContent(Element e, SignedDoc doc) throws DigiDocException {
		Content res=null;
		
		String name=e.getLocalName();
		String id=e.getAttributeValue("Id");
		String mimetype=e.getAttributeValue("MimeType");
		String encoding=e.getAttributeValue("Encoding");
		String algorithm=e.getAttributeValue("Algorithm");
		byte[] data=null;
		int count=e.getChildCount();
		if (count==0) {
			throw new DigiDocException(0,"Invalid content to parse. No CONTENT found.",null);
		}else if (count==1) {
			String xml=e.getChild(0).toXML();
			try {
				data=xml.getBytes(encoding!=null?encoding:"");
			}catch(IOException e2) {
				if(encoding.equals(SignedDoc.BASE64_ENCODING)){
					data=xml.getBytes();
				} else {
					throw new DigiDocException(0,"Error parsing content. Invalid encoding.",e2);
				}
			}
		}else {
			throw new DigiDocException(0,"Number of children > 1. Only one element of data content allowed.",null);
		}
		res=new Content(name,id,doc,data,mimetype,algorithm);
		res.setDataEncoding(encoding);
		return res;
	}
	
	public void setHashAlgortihm(String algorithmURL) throws DigiDocException{
		try{
			URL urlHashAlg = new URL(algorithmURL);
			hashAlgorithm = urlHashAlg.toString();
			createElement();
		} catch (Exception e) {
			throw new DigiDocException(0, "El algoritmo de hash no es una URL vlida", e);
		}
	}

	public String getHashAlgortihm(){
		return hashAlgorithm;		
	}
}