/*
 * KeyInfo.java
 * PROJECT: JDigiDoc
 * DESCRIPTION: Digi Doc functions for creating
 *	and reading signed documents. 
 * AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
 *==================================================
 * Copyright (C) AS Sertifitseerimiskeskus
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * GNU Lesser General Public Licence is available at
 * http://www.gnu.org/copyleft/lesser.html
 *==================================================
 
 */

package ee.sk.digidoc;
import java.io.Serializable;
import java.math.BigInteger;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.util.ArrayList;

import nu.xom.Attribute;
import nu.xom.Element;
import ee.sk.digidoc.factory.CanonicalizationFactory;
import ee.sk.digidoc.utils.CertUtils;
import ee.sk.digidoc.utils.XMLObject;


/**
 * Models the KeyInfo block of an XML-DSIG 
 * signature. In DigiDoc library the key info 
 * allways contains only one subject certificate,
 * e.g. no uplinks and the smaller items like
 * RSA public key modulus and export are not
 * kept separately but calculated online from the
 * signers certificate. That means they are read-only
 * attributes.
 * @author  Veiko Sinivee
 * @version 1.0
 */
public class KeyInfo extends XMLObject implements Serializable
{
	/** parent object - Signature ref */
	private Signature m_signature;
	/** id for URI referenced sign*/
	private String m_id;
	
	private ArrayList m_components=null;
	
	
	public static String LOCALNAME="KeyInfo";
	/** 
	 * Creates new KeyInfo 
	 */
	public KeyInfo() {
		m_signature = null;
		m_components=new ArrayList();
		try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
	}
	
	/** 
	 * Creates new KeyInfo 
	 * @param cert signers certificate
	 */
	public KeyInfo(X509Certificate cert) 
	throws DigiDocException
	{
		setSignersCertificate(cert);
		createElement();
	}
	
	/** 
	 * Creates new KeyInfo 
	 * @param cert signers certificate
	 */
	public KeyInfo(Signature sig,X509Certificate cert) 
	throws DigiDocException
	{
		m_signature=sig;
		setSignersCertificate(cert);
//		createElement();
	}
	
	
	/**
	 * Accessor for Signature attribute
	 * @return value of Signature attribute
	 */
	public Signature getSignature()
	{
		return m_signature;
	}
	
	/**
	 * Mutator for Signature attribute
	 * @param uprops value of Signature attribute
	 */
	public void setSignature(Signature sig)
	{
		m_signature = sig;
		try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
	}
	
	/**
	 * Accessor for signersCert attribute
	 * @return value of signersCert attribute
	 */
	public X509Certificate getSignersCertificate() {
		X509Certificate cert = null;
		if(m_signature != null) {
			CertValue cval = m_signature.getCertValueOfType(CertValue.CERTVAL_TYPE_SIGNER);
			if(cval != null) {
				cert = cval.getCert();
			}
		}
		return cert;
	}
	
	/**
	 * return certificate owners first name
	 * @return certificate owners first name or null
	 */
	public String getSubjectFirstName() {
		X509Certificate cert = getSignersCertificate();
		if(cert != null)
			return CertUtils.getSubjectFirstName(cert);
		else
			return null;
	}
	
	/**
	 * return certificate owners last name
	 * @return certificate owners last name or null
	 */
	public String getSubjectLastName() {
		X509Certificate cert = getSignersCertificate();
		if(cert != null)
			return CertUtils.getSubjectLastName(cert);
		else
			return null;
	}
	
	/**
	 * return certificate owners personal code
	 * @return certificate owners personal code or null
	 */
	public String getSubjectPersonalCode() {
		X509Certificate cert = getSignersCertificate();
		if(cert != null)
			return CertUtils.getSubjectPersonalCode(cert);
		else
			return null;
	}
	
	/**
	 * Mutator for signersCert attribute
	 * @param cert new value for signersCert attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setSignersCertificate(X509Certificate cert) 
	throws DigiDocException
	{
		DigiDocException ex = validateSignersCertificate(cert);
		if(ex != null)
			throw ex;
		if(m_signature != null) {
			CertValue cval = m_signature.getOrCreateCertValueOfType(CertValue.CERTVAL_TYPE_SIGNER);
			cval.setCert(cert);
			createElement();
		}
	}
	
	/**
	 * Helper method to validate a signers cert
	 * @param cert input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateSignersCertificate(X509Certificate cert)
	{
		DigiDocException ex = null;
		if(cert == null)
			ex = new DigiDocException(DigiDocException.ERR_SIGNERS_CERT, 
					"Signers certificate is required", null);
		return ex;
	}
	
	/**
	 * return the signers certificates key modulus
	 * @return signers certificates key modulus
	 */
	public BigInteger getSignerKeyModulus()
	{
		X509Certificate cert =null;
		cert=getSignersCertificate();
		if(cert != null)
			return ((RSAPublicKey)cert.getPublicKey()).getModulus();
		else
			return null;
	}
	
	/**
	 * return the signers certificates key exponent
	 * @return signers certificates key exponent
	 */
	public BigInteger getSignerKeyExponent()
	{
		X509Certificate cert = getSignersCertificate();
		if(cert != null)
			return ((RSAPublicKey)cert.getPublicKey()).getPublicExponent();
		else
			return null;
	}
	
	/**
	 * Helper method to validate the whole
	 * KeyInfo object
	 * @return a possibly empty list of DigiDocException objects
	 */
	public ArrayList validate()
	{
		ArrayList errs = new ArrayList();
		DigiDocException ex = null;
		X509Certificate cert = getSignersCertificate();
		if(cert != null)
			ex = validateSignersCertificate(cert);
		if(ex != null)
			errs.add(ex);        
		return errs;
	}
	
	public void setId(String str) throws DigiDocException{
		DigiDocException ex = validateId(str, false);
		if(ex != null)
			throw ex;
		m_id = str;
		createElement();
		
	}
	
	public String getId(){
		return m_id;
	}
	/** Helper method to validate an id
	 * @param str input data
	 * @param bStrong flag that specifies if Id atribute value is to
	 * be rigorously checked (according to digidoc format) or only
	 * as required by XML-DSIG
	 * @return exception or null for ok
	 */
	private DigiDocException validateId(String str, boolean bStrong)
	{
		DigiDocException ex = null;
		if(str == null)
			ex = new DigiDocException(DigiDocException.ERR_DATA_FILE_ID, 
					"Id is a required attribute", null);
		if(str != null && bStrong && 
				(str.charAt(0) != 'S' || !Character.isDigit(str.charAt(1))))
			ex = new DigiDocException(DigiDocException.ERR_DATA_FILE_ID, 
					"Id attribute value has to be in form S<number>-KeyInfo", null);
		return ex;
	}
	
	
	/**
	 * Accessor for digest attribute
	 * @return value of digest attribute
	 */
	public byte[] getDigest() 
	throws DigiDocException
	{       
		return calculateDigest();
	}
	
	public byte[] calculateDigest()throws DigiDocException{
		CanonicalizationFactory canFac=SignedDoc.getCanonicalizeInstance();
		byte[] tmp = canFac.canonicalize(thisObject, SignedDoc.CANONICALIZATION_METHOD_20010315);
		byte[] hash=SignedDoc.digest(tmp);
		return hash;   
	}
	
	protected Element createElement() throws DigiDocException{
		if(thisObject==null)
			thisObject=makeEasyElement("KeyInfo",SignedDoc.XMLNS_XMLDSIG);
		else{
			if(this.m_signature.isExternalSignature())
				return thisObject;
		}
		if (this.m_id!=null&&!m_id.trim().equals("")){
			thisObject.addAttribute(new Attribute("Id",m_id));
		}
		BigInteger m=getSignerKeyModulus();
		BigInteger e=getSignerKeyExponent();
		if (thisObject.getChildCount()>0)
			thisObject.removeChildren();
		if (m!=null&&
				e!=null){
//			if (thisObject.getChildElements("KeyValue",SignedDoc.XMLNS_XMLDSIG).size()>0){
//			thisObject.removeChild(thisObject.getChildElements("ds:KeyValue").get(0));
//			}
			Element keyValue=makeEasyElement("KeyValue",SignedDoc.XMLNS_XMLDSIG);
			thisObject.appendChild(keyValue);
			
			Element rsaKeyValue=makeEasyElement("RSAKeyValue",SignedDoc.XMLNS_XMLDSIG);
			keyValue.appendChild(rsaKeyValue);
			
			Element modulus=makeEasyElement("Modulus",SignedDoc.XMLNS_XMLDSIG);
			modulus.appendChild(Base64Util.encode(m.toByteArray(), 64));
			
			Element exponent=makeEasyElement("Exponent",SignedDoc.XMLNS_XMLDSIG);
			exponent.appendChild(Base64Util.encode(e.toByteArray(), 64));
			rsaKeyValue.appendChild(modulus);
			rsaKeyValue.appendChild(exponent);
		}
		
		if(m_signature != null){
//			if (thisObject.getChildElements("X509Data",SignedDoc.XMLNS_XMLDSIG).size()>0){
//			thisObject.removeChild(thisObject.getChildElements("ds:X509Data").get(0));
//			}
			CertValue cval = m_signature.getCertValueOfType(CertValue.CERTVAL_TYPE_SIGNER);
			if (cval!=null){
				Element x509Data=makeEasyElement("X509Data",SignedDoc.XMLNS_XMLDSIG);
				Element x509Cert=cval.createElement();
				if (x509Cert.getParent()!=null)
					x509Cert.detach();
				x509Data.appendChild(x509Cert);
				thisObject.appendChild(x509Data);
			}
		}
		
		return thisObject;
	}
	
	protected Element createElement(String namespace) throws DigiDocException{
		return createElement();
	}
	
//	public static KeyInfo parseElement(Node n, Signature sig) {
//		Element e=(Element)n;
//		KeyInfo ki=null;
//		if (e!=null&&e.getLocalName().equals(LOCALNAME)){
////			Elements children=e.getChildElements();
////			ki=new KeyInfo();
////			ki.setSignature(sig);
////			for (int i=0;i<children.size();i++){
////				Element current=children.get(i);
////				if (current.getLocalName().equals("KeyName")){
////					String name=current.getChild(0).getValue();
////					ki.m_components.add(new KeyName(name));
////				}else if (current.getLocalName().equals("KeyValue")){
////					Elements keyValues=current.getChildElements();
////					for (int j=0;j<keyValues.size();j++){
////						KeyValue kv=new KeyValue();
////						Element keyValue=keyValues.get(j);
////						if (keyValue.getLocalName().equals("RSAKeyValue")){
////							String name=current.getChild(0).getValue();
////							ki.m_components.add(new KeyName(name));
////						}else{
////							throw new DigiDocException(DigiDocException.ERR_PARSER_WRONG_ELEMENT,"Only RSAKeyValue element allowed.",null);
////						}
////					}
////				}else if (current.getLocalName().equals("RetrievalMethod")){
////					String uri=current.getAttribute("URI").getValue();
////					RetrievalMethod rm=new RetrievalMethod(uri);
////					Elements transforms=current.getChildElements();
////					for (int j=0;j<transforms.size();j++){
////						Element t=transforms.get(j);
////						t.getAttribute("Algorithm").getValue();
////					}
////					ki.m_components.add(rm);
////				}else if (current.getLocalName().equals("X509Data")){
////					Elements keyValues=current.getChildElements();
////					for (int j=0;j<keyValues.size();j++){
////						KeyValue kv=new KeyValue();
////						Element keyValue=keyValues.get(j);
////						if (keyValue.getLocalName().equals("RSAKeyValue")){
////							String name=current.getChild(0).getValue();
////							ki.m_components.add(new KeyName(name));
////						}else{
////							throw new DigiDocException(DigiDocException.ERR_PARSER_WRONG_ELEMENT,"Only RSAKeyValue element allowed.",null);
////						}
////					}
////				}else if (current.getLocalName().equals("X509Data")){{
////					
////				}
//			ki.thisObject=(Element) n;
//			}
//			
//		return ki;
//	}
	
//	private abstract class AKeyInfoComponent extends XMLObject{}
//	private class KeyValue extends AKeyInfoComponent{
//		public class RSAKeyValue extends XMLObject{
//			BigInteger m_modulus=null;
//			BigInteger m_exponent=null;
//			
//			public RSAKeyValue(){;}
//			
//			public RSAKeyValue(BigInteger modulus,BigInteger exponent){
//				m_modulus=modulus;
//				m_exponent=exponent;				
//			}
//
//			protected Element createElement() throws DigiDocException {
//				if (m_modulus!=null&&m_exponent!=null){
//					thisObject=makeEasyElement("RSAKeyValue",XMLNS_XMLDSIG);
//					
//					Element mod=makeEasyElement("Modulus",XMLNS_XMLDSIG);
//					mod.appendChild(Base64Util.encode(m_modulus.toByteArray()));
//					thisObject.appendChild(mod);
//					
//					Element exp=makeEasyElement("Exponent",XMLNS_XMLDSIG);
//					exp.appendChild(Base64Util.encode(m_exponent.toByteArray()));
//					thisObject.appendChild(exp);
//					return thisObject;
//				}
//				return null;
//			}
//
//			protected Element createElement(String namespace) throws DigiDocException {
//				return createElement();
//			}
//		}
//
//		private ArrayList rsaKeys=null;
//		public KeyValue(){
//			rsaKeys=new ArrayList();
//		}
//		public KeyValue(RSAKeyValue key){
//			rsaKeys=new ArrayList();
//			rsaKeys.add(key);
//		}
//		
//		protected Element createElement() throws DigiDocException {
//			thisObject=makeEasyElement("KeyValue",XMLNS_XMLDSIG);
//			for (int i=0;i<rsaKeys.size();i++){
//				Element rsaKey=((RSAKeyValue)rsaKeys.get(i)).createElement();
//				thisObject.appendChild(rsaKey);
//			}
//			return thisObject;
//		}
//
//		protected Element createElement(String namespace) throws DigiDocException {
//			return createElement();
//		}
//	}
//	private class KeyName extends AKeyInfoComponent{
//		private String m_name=null;
//		public KeyName(){;}
//		public KeyName(String name){
//			m_name=name;
//		}
//		
//		protected Element createElement() throws DigiDocException {
//			if(m_name!=null){
//				thisObject=makeEasyElement("KeyName",XMLNS_XMLDSIG);
//				thisObject.appendChild(m_name);
//				return thisObject;
//			}
//			return null;
//		}
//
//		protected Element createElement(String namespace) throws DigiDocException {
//			return createElement();
//		}
//	}
//	private class RetrievalMethod extends AKeyInfoComponent{
//		private String m_uri=null;
//		private ArrayList m_transforms=null;
//		public RetrievalMethod(){
//			m_transforms=new ArrayList();
//		}
//		public RetrievalMethod(String uri){
//			m_uri=uri;
//		}
//		public void addTransform(String transform){
//			m_transforms.add(transform);
//		}
//		protected Element createElement() throws DigiDocException {
//			if (m_uri!=null){
//				thisObject=makeEasyElement("RetrievalMethod",XMLNS_XMLDSIG);
//				thisObject.addAttribute(new Attribute("URI",m_uri));
//				for (int i=0;i<m_transforms.size();i++){
//					Element transform=makeEasyElement("Transform",SignedDoc.XMLNS_XMLDSIG);
//					transform.addAttribute(new Attribute("Algorithm",(String)m_transforms.get(i)));
//					thisObject.appendChild(transform);
//				}
//				return thisObject;
//			}
//			return null;
//		}
//		protected Element createElement(String namespace) throws DigiDocException {
//			return createElement();
//		}
//		
//	}
//	private class X509Data extends AKeyInfoComponent{
//		private ArrayList m_x509IssuerSerials;
//		private ArrayList m_x509SubjectNames;
//		private ArrayList m_x509Certificates;
////		private ArrayList m_x509CRLs;
//		private ArrayList m_x509SKIs;
//		
//		public X509Data() {
//			m_x509IssuerSerials=new ArrayList();
//			m_x509SubjectNames=new ArrayList();
//			m_x509Certificates=new ArrayList();
////			m_x509CRLs=new ArrayList();
//			m_x509SKIs=new ArrayList();
//		}
//		
//		public void addIssuerSerial(String issuer,BigInteger serial){
//			m_x509IssuerSerials.add(new X509IssuerSerial(issuer,serial));
//		}
//		public void addCertificate(byte[] encoded){
//			m_x509Certificates.add(new X509Certificate(encoded));
//		}
//		public void addSKI(String ski){
//			m_x509SKIs.add(new X509SKI(ski));
//		}
//		public void addSubjectName(String name){
//			m_x509SubjectNames.add(new X509SubjectName(name));
//		}
//		
//		public class X509IssuerSerial extends XMLObject{
//			private String m_issuer=null;
//			private BigInteger m_serial=null;
//			
//			public X509IssuerSerial(){;}
//			public X509IssuerSerial(String issuer,BigInteger serial){
//				m_issuer=issuer;
//				m_serial=serial;
//			}
//			
//			public void setIssuer(String issuer) {
//				this.m_issuer = issuer;
//			}
//			public void setSerial(BigInteger serial) {
//				this.m_serial = serial;
//			}
//			protected Element createElement() throws DigiDocException {
//				if (m_issuer!=null&&m_serial!=null){
//					thisObject=makeEasyElement("X509IssuerSerial",XMLNS_XMLDSIG);
//					Element issuer=makeEasyElement("X509IssuerName",XMLNS_XMLDSIG);
//					issuer.appendChild(m_issuer);
//					thisObject.appendChild(issuer);
//					Element serial=makeEasyElement("X509SerialNumber",XMLNS_XMLDSIG);
//					serial.appendChild(m_serial.toString());
//					thisObject.appendChild(serial);
//					return thisObject;
//				}
//				return null;
//			}
//			protected Element createElement(String namespace) throws DigiDocException {
//				return createElement();
//			}
//		}
//		
//		public class X509SubjectName extends XMLObject{
//			private String m_name=null;
//			
//			public X509SubjectName(){;}
//			public X509SubjectName(String name){
//				m_name=name;
//			}
//			
//			public void setName(String subjectName) {
//				this.m_name = subjectName;
//			}
//			protected Element createElement() throws DigiDocException {
//				if (m_name!=null){
//					thisObject=makeEasyElement("X509SubjectName",XMLNS_XMLDSIG);
//					thisObject.appendChild(m_name);
//					return thisObject;
//				}
//				return null;
//			}
//			protected Element createElement(String namespace) throws DigiDocException {
//				return createElement();
//			}
//			
//
//		}
//
//		public class X509SKI extends XMLObject{
//			private String m_subjectKeyId=null;
//			
//			public X509SKI(){;}
//			public X509SKI(String id){
//				m_subjectKeyId=id;
//			}
//			
//			public void setSubjectKeyId(String id) {
//				this.m_subjectKeyId = id;
//			}
//			protected Element createElement() throws DigiDocException {
//				if (m_subjectKeyId!=null){
//					thisObject=makeEasyElement("X509SKI",XMLNS_XMLDSIG);
//					thisObject.appendChild(m_subjectKeyId);
//					return thisObject;
//				}
//				return null;
//			}
//			protected Element createElement(String namespace) throws DigiDocException {
//				return createElement();
//			}
//			
//
//		}
//
//		public class X509Certificate extends XMLObject{
//			private byte[] m_encoded=null;
//			
//			public X509Certificate(){;}
//			public X509Certificate(String b64){
//				m_encoded=Base64Util.decode(b64);
//			}
//			public X509Certificate(byte[] encoded){
//				m_encoded=encoded;
//			}
//			
//			public void setX509Cert(byte[] encoded) {
//				m_encoded=encoded;
//			}
//			protected Element createElement() throws DigiDocException {
//				if (m_encoded!=null){
//					thisObject=makeEasyElement("X509Certificate",XMLNS_XMLDSIG);
//					thisObject.appendChild(Base64Util.encode(m_encoded));
//					return thisObject;
//				}
//				return null;
//			}
//			protected Element createElement(String namespace) throws DigiDocException {
//				return createElement();
//			}
//			
//
//		}
//
//		protected Element createElement() throws DigiDocException {
//			if (m_x509Certificates.size()>0||
//					m_x509IssuerSerials.size()>0||
//					m_x509SKIs.size()>0||
//					m_x509SubjectNames.size()>0){
//				thisObject=makeEasyElement("X509Data",XMLNS_XMLDSIG);
//				for (int i=0;i<m_x509Certificates.size();i++){
//					Element cert=((X509Certificate)m_x509Certificates.get(i)).createElement();
//					thisObject.appendChild(cert);
//				}
//				for (int i=0;i<m_x509IssuerSerials.size();i++){
//					Element isn=((X509IssuerSerial)m_x509IssuerSerials.get(i)).createElement();
//					thisObject.appendChild(isn);
//				}
//				for (int i=0;i<m_x509SKIs.size();i++){
//					Element ski=((X509SKI)m_x509SKIs.get(i)).createElement();
//					thisObject.appendChild(ski);
//				}
//				for (int i=0;i<m_x509SubjectNames.size();i++){
//					Element sn=((X509SubjectName)m_x509SubjectNames.get(i)).createElement();
//					thisObject.appendChild(sn);
//				}
//				return thisObject;
//			}
//			return null;
//		}
//
//		protected Element createElement(String namespace) throws DigiDocException {
//			// TODO Auto-generated method stub
//			return null;
//		}
//
//	}
}
