
 /* Reference.java
 * PROJECT: JDigiDoc
 * DESCRIPTION: Digi Doc functions for creating
 *	and reading signed documents. 
 * AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
 *==================================================
 * Copyright (C) AS Sertifitseerimiskeskus
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * GNU Lesser General Public Licence is available at
 * http://www.gnu.org/copyleft/lesser.html
 *==================================================
 *
 * THIS CLASS HAS BEEN MODIFIED:
 * -----------------------------
 * MODIFICATION: Canonicalization method is always published, if not specified
 * xml-c14n is used. Digests have not 'newline' characters.
 * AUTHOR: Ministerio de Administraciones Pblicas. Gobierno de Espaa.
 */

package ee.sk.digidoc;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.MultipleParentException;
import ee.sk.digidoc.utils.UUIDUtils;
import ee.sk.digidoc.utils.XMLObject;


/**
 * Represents and XML-DSIG reference block
 * that referrs to a particular piece of
 * signed XML data and contains it's hash code.
 * @author  Veiko Sinivee
 * @version 1.0
 */
public class Reference extends XMLObject implements Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private static final String LOCALNAME = "Reference";
	/** reference to parent SignedInfo object */
	private SignedInfo m_sigInfo;
	/** URI to signed XML data */
	private String m_uri;
	/** selected digest algorithm */
	private String m_digestAlgorithm;
	/** digest data */
	private byte[] m_digestValue;
	/** transform algorithm */
	private ArrayList m_transformsAlgorithm;
	
	private String m_type;
	
	private String m_id;
	
	private HashMap m_transElem = new HashMap();
	
	/** 
	 * Creates new Reference. Initializes
	 * everything to null
	 * @param sigInfo reference to parent SignedInfo object
	 */
	public Reference(SignedInfo sigInfo) 
	{
		m_sigInfo = sigInfo;
		m_type=null;
		m_uri = null;
		m_digestAlgorithm = null;
		m_digestValue = null;
		m_transformsAlgorithm = new ArrayList();
		m_transElem = new HashMap();
		m_id=null;
		try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
	}
	
	/** 
	 * Creates new Reference 
	 * @param sigInfo reference to parent SignedInfo object
	 * @param uri reference uri pointing to signed XML data
	 * @param algorithm sigest algorithm identifier
	 * @param digest message digest data
	 * @param transform transform algorithm
	 * @throws DigiDocException for validation errors
	 */
	public Reference(SignedInfo sigInfo, String uri, String algorithm, 
			byte[] digest, String[] transforms) 
	throws DigiDocException
	{
		m_sigInfo = sigInfo;
		setUri(uri);
		setDigestAlgorithm(algorithm);
		setDigestValue(digest);
		m_transformsAlgorithm = new ArrayList();
		m_transElem = new HashMap();
		m_id=LOCALNAME+"-"+UUIDUtils.generateUUID();
		if (transforms!=null)
			for (int i=0;i<transforms.length;i++)
				addTransformAlgorithm(transforms[i]);
		createElement();
	}
	
	public Reference(SignedInfo sigInfo, String uri,String type, String algorithm, 
			byte[] digest, String[] transforms) 
	throws DigiDocException
	{
		m_sigInfo = sigInfo;
		setUri(uri);
		setType(type);
		setDigestAlgorithm(algorithm);
		setDigestValue(digest);
		m_transformsAlgorithm = new ArrayList();
		m_transElem = new HashMap();
		if (transforms!=null)
			for (int i=0;i<transforms.length;i++)
				addTransformAlgorithm(transforms[i]);
		m_id=LOCALNAME+"-"+UUIDUtils.generateUUID();
		createElement();
	}
	
	/** 
	 * Creates new Reference 
	 * and initializes it with default
	 * values from the DataFile
	 * @param sigInfo reference to parent SignedInfo object
	 * @param df DataFile object
	 * @throws DigiDocException for validation errors
	 */
	public Reference(SignedInfo sigInfo, Content df) 
	throws DigiDocException{
		m_sigInfo = sigInfo;
		String mode=m_sigInfo.getSignature().getSignedDoc().getSignatureMode();
		if (mode.equals(SignedDoc.MODE_ENVELOPED_SIGNATURE)
				||df.getId()==null
				||df.getId().length()==0)
			setUri("");
		else
			setUri("#" + df.getId());
		m_transformsAlgorithm = new ArrayList();
		m_transElem = new HashMap();
		m_type=null;
		setDigestAlgorithm(SignedDoc.SHA1_DIGEST_ALGORITHM);
		setDigestValue(df.getDigest());
		m_id=LOCALNAME+"-"+UUIDUtils.generateUUID();
		createElement();		
	}
	
	/**
	 * Accessor for sigInfo attribute
	 * @return value of sigInfo attribute
	 */
	public SignedInfo getSignedInfo() {
		return m_sigInfo;
	}
	
	/**
	 * Mutator for sigInfo attribute
	 * @param sigInfo new value for sigInfo attribute
	 */    
	public void setSignedInfo(SignedInfo sigInfo) 
	{
		m_sigInfo = sigInfo;
		try {
			createElement();
		} catch (DigiDocException e) {
			;
		}
	}
	
	/** 
	 * Creates new Reference 
	 * and initializes it with default
	 * values from the SignedProperties
	 * @param sigInfo reference to parent SignedInfo object
	 * @param sp SignedProperties object
	 * @throws DigiDocException for validation errors
	 */
	public Reference(SignedInfo sigInfo, SignedProperties sp) 
	throws DigiDocException
	{
		m_sigInfo = sigInfo;
		setUri(sp.getTarget() + "-SignedProperties");
		setDigestAlgorithm(SignedDoc.SHA1_DIGEST_ALGORITHM);
		setDigestValue(sp.calculateDigest());
		m_transformsAlgorithm = new ArrayList();
//		addTransformAlgorithm(SignedDoc.CANONICALIZATION_METHOD_20010315);
		createElement();
		
	}
	
	/** Creates new Reference 
	 * and initializes it with default
	 * values from the SignedProperties
	 * @param sigInfo reference to parent SignedInfo object
	 * @param sp SignedProperties object
	 * @throws DigiDocException for validation errors
	 */
	public Reference(SignedInfo sigInfo, KeyInfo ki) 
	throws DigiDocException
	{
		m_sigInfo = sigInfo;
		setUri("#"+ki.getId());
		setDigestAlgorithm(SignedDoc.SHA1_DIGEST_ALGORITHM);
		setDigestValue(ki.calculateDigest());
		m_transformsAlgorithm = new ArrayList();
		m_id=LOCALNAME+"-"+UUIDUtils.generateUUID();
		createElement();
		
	}
	
	
	/**
	 * Accessor for uri attribute
	 * @return value of uri attribute
	 */
	public String getUri() {
		return m_uri;
	}
	
	/**
	 * Mutator for uri attribute
	 * @param str new value for uri attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setUri(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateUri(str);
		if(ex != null)
			throw ex;
		m_uri = str;
	}
	
	/**
	 * Helper method to validate a uri
	 * @param str input data
	 * @return exception or null for ok
	 */
	
	private DigiDocException validateUri(String str)
	{
		DigiDocException ex = null;
		if(str == null)
			ex = new DigiDocException(DigiDocException.ERR_REFERENCE_URI, 
					"URI cannot be null", null);
		return ex;
	}
	
	/**
	 * Accessor for digestAlgorithm attribute
	 * @return value of digestAlgorithm attribute
	 */
	public String getDigestAlgorithm() {
		return m_digestAlgorithm;
	}
	
	/**
	 * Mutator for digestAlgorithm attribute
	 * @param str new value for digestAlgorithm attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setDigestAlgorithm(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateDigestAlgorithm(str);
		if(ex != null)
			throw ex;
		m_digestAlgorithm = str;
	}
	
	/**
	 * Helper method to validate a digest algorithm
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateDigestAlgorithm(String str)
	{
		DigiDocException ex = null;
		if(str == null || !str.equals(SignedDoc.SHA1_DIGEST_ALGORITHM))
			ex = new DigiDocException(DigiDocException.ERR_DIGEST_ALGORITHM, 
					"Currently supports only SHA1 digest algorithm", null);
		return ex;
	}
	
	/**
	 * Accessor for digestValue attribute
	 * @return value of digestValue attribute
	 */
	public byte[] getDigestValue() {
		return m_digestValue;
	}
	
	/**
	 * Mutator for digestValue attribute
	 * @param data new value for digestValue attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setDigestValue(byte[] data) 
	throws DigiDocException
	{
		DigiDocException ex = validateDigestValue(data);
		if(ex != null)
			throw ex;
		m_digestValue = data;
	}
	
	/**
	 * Helper method to validate a digest value
	 * @param data input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateDigestValue(byte[] data)
	{
		DigiDocException ex = null;
		if(data == null || data.length != SignedDoc.SHA1_DIGEST_LENGTH)
			ex = new DigiDocException(DigiDocException.ERR_DIGEST_LENGTH, 
					"SHA1 digest data is allways "+SignedDoc.SHA1_DIGEST_LENGTH+" bytes of length", null);
		return ex;
	}
	
	/**
	 * Accessor for transformAlgorithm attribute
	 * @return value of transformAlgorithm attribute
	 */
	public String[] getTransformAlgorithm() {
		return (String[])m_transformsAlgorithm.toArray();
	}
	
	/**
	 * Mutator for transformAlgorithm attribute.
	 * Currently supports only one transform which
	 * has to be digidoc detatched document transform
	 * or none at all.
	 * @param str new value for transformAlgorithm attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void addTransformAlgorithm(String str) 
	throws DigiDocException
	{
		if (str!=null&&str.length()>0){
			if (!m_transformsAlgorithm.contains(str)){
				m_transformsAlgorithm.add(str);
				createElement();
			}
		}else
			throw new DigiDocException(DigiDocException.ERR_TRANSFORM_ALGORITHM,"Null or size 0 transformAlgorithm error.",null);
	}
	
	/**
	 * Mutator for transformAlgorithm attribute.
	 * Currently supports only one transform which
	 * has to be digidoc detatched document transform
	 * or none at all.
	 * @param str new value for transformAlgorithm attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void addTransformAlgorithm(String str, String tag, String ns, String value) 
	throws DigiDocException
	{
		if (str!=null&&str.length()>0 && tag!=null&&tag.length()>0 && ns!=null&&ns.length()>0 && value!=null&&value.length()>0){
			if (!m_transformsAlgorithm.contains(str)){
				m_transformsAlgorithm.add(str);	
				Element elem = makeEasyElement(tag,ns);
				elem.appendChild(value);
				m_transElem.put(str,elem);
				createElement();
			}		
		}else
			throw new DigiDocException(DigiDocException.ERR_TRANSFORM_ALGORITHM,"Null or size 0 transformAlgorithm error.",null);
	}
	
	/**
	* Helper method to validate a transform algorithm
	* @param str input data
	* @return exception or null for ok
	*/
	private DigiDocException validateTransformAlgorithm(ArrayList str)
	{
		DigiDocException ex = null;
		return ex;
	}
	
	/**
	 * Helper method to validate the whole
	 * Reference object
	 * @return a possibly empty list of DigiDocException objects
	 */
	public ArrayList validate()
	{
		ArrayList errs = new ArrayList();
		DigiDocException ex = validateUri(m_uri);
		if(ex != null)
			errs.add(ex);
		ex = validateDigestAlgorithm(m_digestAlgorithm);
		if(ex != null)
			errs.add(ex);
		ex = validateDigestValue(m_digestValue);
		if(ex != null)
			errs.add(ex);
		ex = validateTransformAlgorithm(m_transformsAlgorithm);
		if(ex != null)
			errs.add(ex);
		return errs;
	}
	
	protected Element createElement(String namespace) throws DigiDocException{
		return createElement();
	}
	
	protected Element createElement() throws DigiDocException{
		if(thisObject==null)
			thisObject=makeEasyElement("Reference",SignedDoc.XMLNS_XMLDSIG);
		
		if(thisObject.getChildCount()>0)
			thisObject.removeChildren();
		if (m_uri!=null&&m_sigInfo!=null){
			if (m_type==null) {//Intentamos adivinar el type
				if(m_uri.indexOf("SignedProperties") != -1) {
					String namespace=SignedDoc.DEFAULT_XMLNS_ETSI;
					String format=m_sigInfo.getSignature().getSignedDoc().getFormat();
					if (format.equals(SignedDoc.FORMAT_XAdES_1_1_1))
						namespace=SignedDoc.XMLNS_ETSI_1_1_1;
					else if (format.equals(SignedDoc.FORMAT_XAdES_1_2_2))
						namespace=SignedDoc.XMLNS_ETSI_1_2_2;
					else if (format.equals(SignedDoc.FORMAT_XAdES_1_3_2))
						namespace=SignedDoc.TYPE_URI_ETSI_1_3_2;
					thisObject.addAttribute(new Attribute("Type",namespace+"SignedProperties"));
				}else if (m_uri.indexOf("KeyInfo") != -1){
					thisObject.addAttribute(new Attribute("Type",SignedDoc.XMLNS_XMLDSIG+"KeyInfo"));
				}
			}else {
				thisObject.addAttribute(new Attribute("Type",m_type));
			}
			thisObject.addAttribute(new Attribute("URI",m_uri));
		}
		
		if (m_id!=null&&m_id.trim().length()>0) {
			thisObject.addAttribute(new Attribute("Id",m_id));
		}
		
		if (m_transformsAlgorithm!=null&&
				m_transformsAlgorithm.size()>0){
			if (thisObject.getChildElements("ds:Transforms").size()>0){
				thisObject.removeChild(thisObject.getChildElements("ds:Transforms").get(0));
			}
			Element transforms=makeEasyElement("Transforms",SignedDoc.XMLNS_XMLDSIG);
			Iterator it=m_transformsAlgorithm.iterator();
			while(it.hasNext()){
				Element transform=makeEasyElement("Transform",SignedDoc.XMLNS_XMLDSIG);
				String str = (String)it.next();
				transform.addAttribute(new Attribute("Algorithm",str));							
				if(m_transElem != null && m_transElem.size() > 0){
					Element trEl = (Element) m_transElem.get(str);
					if(trEl != null && m_transElem.containsKey((String) str)){										
						try{
							transform.appendChild(trEl);
						}
						catch(MultipleParentException e){
							trEl = (Element) trEl.copy();
							transform.appendChild(trEl);
						}
					}
				}				
				transforms.appendChild(transform);				
			}
			thisObject.appendChild(transforms);
		}
		if (m_digestAlgorithm!=null){
			if (thisObject.getChildElements("ds:DigestMethod").size()>0){
				thisObject.removeChild(thisObject.getChildElements("ds:DigestMethod").get(0));
			}
			Element digestMethod=makeEasyElement("DigestMethod",SignedDoc.XMLNS_XMLDSIG);
			digestMethod.addAttribute(new Attribute("Algorithm",m_digestAlgorithm));
			thisObject.appendChild(digestMethod);
		}
		if (m_digestValue!=null){
			if (thisObject.getChildElements("ds:DigestValue").size()>0){
				thisObject.removeChild(thisObject.getChildElements("ds:DigestValue").get(0));
			}
			Element digestValue=makeEasyElement("DigestValue",SignedDoc.XMLNS_XMLDSIG);
			digestValue.appendChild(Base64Util.encode(m_digestValue, m_digestValue.length));
			thisObject.appendChild(digestValue);
		}
		
		return thisObject;
	}

	public String getType() {
		return m_type;
	}

	public void setType(String type) {
		this.m_type = type;
	}
	
	public String getId() {
		return m_id;
	}

//	public static Reference parseElement(Node n, SignedInfo si) throws DigiDocException {
//		Element e=(Element)n;
//		Reference ref=null;
//		if (e!=null&&e.getLocalName().equals(LOCALNAME)){
//			//Canonicalization Method
//			String uri=e.getAttributeValue("URI");
//			Elements t=e.getChildElements("Transforms",XMLNS_XMLDSIG);
//			if (t!=null&&t.size()>0){
//				Element transformsElement=t.get(0);
//				Elements transforms=transformsElement.getChildElements();
//				String[] algorithms=new String[transforms.size()];
//				for (int i=0;i<transforms.size();i++){
//					Element trans=transforms.get(i);
//					String algorithm=trans.getAttributeValue("Algorithm");
//					algorithms[i]=algorithm;
//				}
//				Element digMeth=e.getChildElements("DigestMethod",XMLNS_XMLDSIG).get(0);
//				String digestMethod=digMeth.getAttributeValue("Algorithm");
//				Element digVal=e.getChildElements("DigestValue",XMLNS_XMLDSIG).get(0);
//				
////				String digestValueStr=digVal.getChildElements().get(0).toString();
//				String digestValueStr=digVal.getChild(0).getValue();
//				byte[] digestValue=Base64Util.decode(digestValueStr);
//				ref=new Reference(si,uri,digestMethod,digestValue,algorithms);
//			}			
//		}else{
//			throw new DigiDocException(DigiDocException.ERR_PARSER_WRONG_ELEMENT,"Null member or wrong localname, expected "+LOCALNAME,null);
//		}
//		return ref;
//	}

}
