/*
 * SignedDoc.java
 * PROJECT: JDigiDoc
 * DESCRIPTION: Digi Doc functions for creating
 *	and reading signed documents. 
 * AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
 *==================================================
 * Copyright (C) AS Sertifitseerimiskeskus
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 * GNU Lesser General Public Licence is available at
 * http://www.gnu.org/copyleft/lesser.html
 *==================================================
 *
 * THIS CLASS HAS BEEN MODIFIED:
 * -----------------------------
 * MODIFICATION: Added possibility of modifying the tag name. Modifications 
 * on the class datafile has been updated here. It is not possible to verify
 * the signature from applet so the feature has been removed in order to reduce
 * libraries size. Certificate reader has been updated. References to xades and 
 * xmldsig namespaces has been updated.
 * AUTHOR: Ministerio de Administraciones Pblicas. Gobierno de Espaa.
 */

package ee.sk.digidoc;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import nu.xom.Attribute;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;
import nu.xom.NoSuchChildException;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ParentNode;
import nu.xom.ParsingException;
import nu.xom.Serializer;
import nu.xom.ValidityException;
import nu.xom.XPathContext;
import nu.xom.canonical.Canonicalizer;

import org.apache.xml.security.transforms.implementations.TransformXPath;

import ee.sk.digidoc.factory.CanonicalizationFactory;
import ee.sk.digidoc.factory.XOMBuilderFactory;
import ee.sk.digidoc.factory.XOMCanonicalizationFactory;
import ee.sk.digidoc.utils.Base64;
import ee.sk.digidoc.utils.UUIDUtils;



/**
 * Represents an instance of signed doc
 * in DIGIDOC format. Contains one or more
 * DataFile -s and zero or more Signature -s.
 * @author  MRC
 * @version 2.0
 */
public class SignedDoc implements Serializable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	/** digidoc format */
	private String m_format;
	/** format version */
	private String m_version;

	/**Datafile object*/
	private Content m_content;
	private Content m_contentEnveloped;
	private DataObjectFormat m_dataObject;
	
	/**CoSignature Members*/
	private static Document m_coSignatureObject=null;
	private static Element m_coSignatureTemp=null;
	private static Signature m_cosign=null;
	
	/** Signature objects */
	private ArrayList m_signatures;
	
	/** @version 3.1.1 
	 * Signature Mode*/
	private String m_mode = MODE_DETACHED_SIGNATURE;	
	private Document m_xml;
	
	public static final String FORMAT_XAdES_1_1_1 				= "XAdES 1.1.1";
	public static final String FORMAT_XAdES_1_2_2 				= "XAdES 1.2.2";
	public static final String FORMAT_XAdES_1_3_2 				= "XAdES 1.3.2";
	public static final String FORMAT_XMLDSIG 					= "SignedDoc";
	
	/*Constants*/
	public static final String SHA1_DIGEST_ALGORITHM			= "http://www.w3.org/2000/09/xmldsig#sha1";
	public static final int    SHA1_DIGEST_LENGTH				= 20;
	public static final String CANONICALIZATION_METHOD_20010315	= Canonicalizer.EXCLUSIVE_XML_CANONICALIZATION;
	public static final String RSA_SHA1_SIGNATURE_METHOD		= "http://www.w3.org/2000/09/xmldsig#rsa-sha1";
	public static final String ENVELOPED_TRANSFORM				= "http://www.w3.org/2000/09/xmldsig#enveloped-signature";
	
	public static final String XMLNS_XMLDSIG 					= "http://www.w3.org/2000/09/xmldsig#";
	public static final String XMLNS_ETSI_1_1_1 				= "http://uri.etsi.org/01903/v1.1.1#";
	public static final String XMLNS_ETSI_1_3_2 				= "http://uri.etsi.org/01903/v1.3.2#";
	public static final String XMLNS_ETSI_1_2_2 				= "http://uri.etsi.org/01903/v1.2.2#";
	public static final String DEFAULT_XMLNS_ETSI 				= XMLNS_ETSI_1_3_2;
	
	public static final String VERSION_1_0 						= "1.0";
	public static final String LIB_NAME 						= "OpenXAdES from DigiDoc dist.";
	public static final String LIB_VERSION						= "5";
	private static final String ROOT_NAME 						= "AFIRMA";
	
	public static final String TYPE_URI_ETSI_1_3_2 				= "http://uri.etsi.org/01903#";
	public static final String BASE64_ENCODING					= "http://www.w3.org/2000/09/xmldsig#base64";
	public static final String DEFAULT_MIMETYPE					= "application/octet-stream";
	
	/**
	 * The signature is over content found within an Object element of the 
	 * signature itself. The Object(or its content) is identified via a 
	 * Reference (via a URI fragment idenitifier or transform).
	 * @version 3.1.1
	 */
	public static final String MODE_ENVELOPING_SIGNATURE 		= "ENVELOPING";
	
	/**<b>The signature is over content external to the Signature element</b>, 
	 * and can be identified via a URI or transform. Consequently, the 
	 * signature is "detached" from the content it signs. <b>This definition 
	 * typically applies to separate data objects, but it also includes 
	 * the instance where the Signature and data object reside within the 
	 * same XML document but are sibling elements.</b> 
	 * We 
	 * 
	 * @version 3.1.1
	 */
	public static final String MODE_DETACHED_SIGNATURE 			= "DETACHED";
	
	/**
	 * The signature is over the XML content that contains the signature as 
	 * an element. The content provides the root XML document element. 
	 * Obviously, enveloped signatures must take care not to include their 
	 * own value in the calculation of the SignatureValue.
	 * <b>(default mode)</b> 
	 * @version 3.1.1
	 */
	public static final String MODE_ENVELOPED_SIGNATURE 		= "ENVELOPED";
	
	private static CanonicalizationFactory canFac = new XOMCanonicalizationFactory();
	
	public static CanonicalizationFactory getCanonicalizeInstance(){
		if (canFac==null)
			canFac=new XOMCanonicalizationFactory();
		return canFac;
	}
	
	/** 
	 * Creates new SignedDoc 
	 * Initializes everything to null
	 */
	public SignedDoc() {
		m_format = null;
		m_version = null;
		m_signatures = null;
		m_xml=null;
		//Default signature mode
		m_mode=MODE_DETACHED_SIGNATURE;
	}
	
	/** 
	 * Creates new SignedDoc 
	 * @param format file format name
	 * @param version file version number
	 * @throws DigiDocException for validation errors
	 */
	public SignedDoc(String format, String version) 
	throws DigiDocException
	{
		setFormat(format);
		setVersion(version);
		m_signatures = null;
		m_xml=null;
		//Default signature mode
		m_mode=MODE_DETACHED_SIGNATURE;
	}
	
	
	/** 
	 * Creates new SignedDoc 
	 * @param format file format name
	 * @param version file version number
	 * @throws DigiDocException for validation errors
	 */
	public SignedDoc(String format, String version,String mode) 
	throws DigiDocException
	{
		setFormat(format);
		setVersion(version);
		m_signatures = null;
		m_xml=null;
		if (mode!=null&&(mode.equalsIgnoreCase(MODE_DETACHED_SIGNATURE)
				|| mode.equals(MODE_ENVELOPED_SIGNATURE)
				|| mode.equals(MODE_ENVELOPING_SIGNATURE)))
			m_mode=mode;
		else
			m_mode=MODE_DETACHED_SIGNATURE;
	}
	
	
	
	/*****************************/
	
	public String getSignatureMode(){
		return m_mode;
	}
	
	public void setSigntatureMode(String mode)throws DigiDocException{
		
		if (mode!=null && (mode.equalsIgnoreCase(MODE_DETACHED_SIGNATURE)
				|| (mode.equals(MODE_ENVELOPED_SIGNATURE))
				|| mode.equals(MODE_ENVELOPING_SIGNATURE))){
			m_mode=mode;
		}
		
		else
			throw new DigiDocException(DigiDocException.ERR_SIGNATURE_MODE,"Invalid signature mode. Only DETACHED, ENVELOPED or ENVELOPING are allowed.",null);
	}
	
	/*******************************/
	
	
	/**
	 * Accessor for format attribute
	 * @return value of format attribute
	 */
	public String getFormat() {
		return m_format;
	}
	
	/**
	 * Mutator for format attribute
	 * @param str new value for format attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setFormat(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateFormat(str);
		if(ex != null)
			throw ex;
		m_format = str;
	}
	
	/**
	 * Helper method to validate a format
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateFormat(String str)
	{
		DigiDocException ex = null;
		if(str == null || 
				(!str.equals(FORMAT_XAdES_1_1_1) 
//						&& !str.equals(FORMAT_XAdES_1_2_2)
						&& !str.equals(FORMAT_XAdES_1_3_2)
						&& !str.equals(FORMAT_XMLDSIG)))
			ex = new DigiDocException(DigiDocException.ERR_DIGIDOC_FORMAT, 
					"Currently supports only XADES 1.1.1, XADES 1.3.2 and XMLDSIGNATURE formats", null);
		return ex;
	}
	
	/**
	 * Accessor for version attribute
	 * @return value of version attribute
	 */
	public String getVersion() {
		return m_version;
	}
	
	/**
	 * Mutator for version attribute
	 * @param str new value for version attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setVersion(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateVersion(str);
		if(ex != null)
			throw ex;
		m_version = str;
	}
	
	/**
	 * Helper method to validate a version
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateVersion(String str)
	{
		DigiDocException ex = null;
		if(str == null || 
				(!str.equals(VERSION_1_0))) 
			ex = new DigiDocException(DigiDocException.ERR_DIGIDOC_VERSION, 
					"Currently supports only versions 1.0", null);
		return ex;
	}
	
	/**
	 * return a new available DataFile id
	 * @retusn new DataFile id
	 */
	public String getNewDataFileId()
	{
		String uid=UUIDUtils.generateUUID();
		String id = "Object-" + uid;
		return id;
	}
	
	public Content setContent(byte[] data, String tagName, String encoding) throws DigiDocException{
		return setContent(data, tagName, null, encoding);
	}
	
	public Content setContent(byte[] data, String tagName, String mimetype, String encoding) throws DigiDocException{
		if (countSignatures()>0){
			throw new DigiDocException(DigiDocException.ERR_SIGATURES_EXIST,"Not allowed to change content after a signature exists.",null);
		}
		if (data==null||data.length==0){
			throw new DigiDocException(DigiDocException.ERR_DATA_FILE_SIZE,"Content size incorrect. At least 1byte data must be to-be-signed",null);
		}
		byte[] realData=parseContent(data);
		Content df = null;		
			df = new Content(tagName, getNewDataFileId(), this, realData, mimetype);
		if(encoding == null){
			df.setDataEncoding("UTF-8");
		}
		else{
			df.setDataEncoding(encoding);
		}
		
		m_content=df;
		return m_content;
	}
	
	public Content setContent(byte[] data, String tagName, String mimetype, String encoding, String algHash) throws DigiDocException{
		if (countSignatures()>0){
			throw new DigiDocException(DigiDocException.ERR_SIGATURES_EXIST,"Not allowed to change content after a signature exists.",null);
		}
		if (data==null||data.length==0){
			throw new DigiDocException(DigiDocException.ERR_DATA_FILE_SIZE,"Content size incorrect. At least 1byte data must be to-be-signed",null);
		}
		byte[] realData=parseContent(data);
		Content df = null;		
			df = new Content(tagName, getNewDataFileId(), this, realData, mimetype);
		if(encoding == null){
			df.setDataEncoding("UTF-8");
		}
		else{
			df.setDataEncoding(encoding);
		}		
		if(algHash != null){
			df.setHashAlgortihm(algHash);
		}
		
		m_content=df;
		return m_content;
	}
	
	public Content setContentEnveloped(byte[] data, String tagName, String mimetype, String encoding) throws DigiDocException, UnsupportedEncodingException{
		if (countSignatures()>0)
			throw new DigiDocException(DigiDocException.ERR_SIGATURES_EXIST,"Not allowed to change content after a signature exists.",null);
		if (data==null||data.length==0)
			throw new DigiDocException(DigiDocException.ERR_DATA_FILE_SIZE,"Content size incorrect. At least 1byte data must be to-be-signed",null);
			
			byte[] realData=parseContent(data);						
			m_contentEnveloped = new Content(tagName, getNewDataFileId(), this, realData, mimetype);
			Element elemEnveloped = (Element) m_contentEnveloped.getElement().copy();
			Nodes ns = null;
	        XPathContext context=XPathContext.makeNamespaceContext(elemEnveloped);
	        context.addNamespace("xades", XMLNS_ETSI_1_3_2);      
	        context.addNamespace("ds", XMLNS_XMLDSIG);
			ns = elemEnveloped.query("//ds:Signature",context);
			for (int i=0 ; i<ns.size(); i++){
				try{
					elemEnveloped.removeChild(ns.get(i));
				} catch (NoSuchChildException e){

				} catch (Exception e){
					System.out.println("Error :" + e.getCause());
				}
			}
			try {
				realData=parseContent(elemEnveloped.toXML().getBytes(encoding));
			} catch (UnsupportedEncodingException e) {
				throw e;
			} catch (DigiDocException e) {
				throw e;
			}	
			
			Content df = null;
			df = new Content(tagName, getNewDataFileId(), this, realData, mimetype);
			if(encoding == null){
				df.setDataEncoding("UTF-8");
			}
			else{
				df.setDataEncoding(encoding);
			}
			
			m_content=df;
		return m_content;
	}
	
	public Content setContent(byte[] data,String tagName, String mimetype,String description,String encoding,String oid,boolean forceDataRef) throws DigiDocException{
		if (countSignatures()>0)
			throw new DigiDocException(DigiDocException.ERR_SIGATURES_EXIST,"Not allowed to change content after a signature exists.",null);
		if (data==null||data.length==0)
			throw new DigiDocException(DigiDocException.ERR_DATA_FILE_SIZE,"Content size incorrect. At least 1byte data must be to-be-signed",null);
		byte[] realData=parseContent(data);
		String id=getNewDataFileId();
		Content df = null;		
		df = new Content(tagName, id, this, realData, mimetype);
		if(encoding == null)
			df.setDataEncoding("UTF-8");
		else
			df.setDataEncoding(encoding);

		m_content=df;
		return m_content;
	}
	
	public Content getContent(){
		return m_content;
	}
	
	/**
	 * return the count of Signature objects
	 * @return count of Signature objects
	 */
	public int countSignatures()
	{
		return ((m_signatures == null) ? 0 : m_signatures.size());
	}
	
	/**
	 * return a new available Signature id
	 * @return new Signature id
	 */
	public String getNewSignatureId()
	{
		String uid=UUIDUtils.generateUUID();
		String id = "signature-" + uid;
		return id;
	}
	
	/**
	 * Adds a new uncomplete signature to signed doc
	 * @param cert signers certificate
	 * @param claimedRoles signers claimed roles
	 * @param adr signers address
	 * @return new Signature object
	 */
	public Signature prepareSignature(X509Certificate cert, String[] claimedRoles, SignatureProductionPlace adr) throws DigiDocException{
		
		Signature sig = new Signature(this);
		sig.setId(getNewSignatureId());
		// create SignedInfo block
		SignedInfo si = new SignedInfo(sig, RSA_SHA1_SIGNATURE_METHOD, CANONICALIZATION_METHOD_20010315);
		// add DataFile reference
		Content df = getContent();
		sig.setContent(df);    
		
		// add reference to dataSign
		Reference ref = new Reference(si, df);
		if (m_mode.equals(MODE_ENVELOPED_SIGNATURE)){
//			ref.addTransformAlgorithm(ENVELOPED_TRANSFORM);
			ref.addTransformAlgorithm(TransformXPath.implementedTransformURI,"XPath",XMLNS_XMLDSIG,"not(ancestor-or-self::ds:Signature)");
			ref.addTransformAlgorithm(CANONICALIZATION_METHOD_20010315);			
			ref.setUri("");
			if(m_contentEnveloped != null){	
				//Tras calcular el hash de los datos sin firmas se vuelve a los datos originales
				sig.setContent(m_contentEnveloped);
				m_content = m_contentEnveloped;
			}
			else{
				throw new DigiDocException(0, "El contenido firmado debe ser Enveloped", null);
			}			
		}else{
			ref.addTransformAlgorithm(CANONICALIZATION_METHOD_20010315);
		}
		si.addReference(ref);
		
		// create key info
		KeyInfo ki = new KeyInfo(sig,cert);
		if (m_format.equals(FORMAT_XMLDSIG)){
			ki.setId(sig.getId()+"-KeyInfo");
		}else{
			ki.setId("");
		}
		sig.setKeyInfo(ki);
		ki.setSignature(sig);
		
		// add referente to keyInfo
		if (m_format.equals(FORMAT_XMLDSIG)){
			Reference refki = new Reference(si, ki);
			refki.addTransformAlgorithm(CANONICALIZATION_METHOD_20010315);
			refki.setUri("#"+ki.getId());
			si.addReference(refki);
		}
		
		// add certID to Signature
		CertID cid = new CertID(sig, cert, CertID.CERTID_TYPE_SIGNER);
		sig.addCertID(cid);
		
		
		// create signed properties		
		if (m_format.equals(FORMAT_XAdES_1_1_1)||
				m_format.equals(FORMAT_XAdES_1_2_2)||
				m_format.equals(FORMAT_XAdES_1_3_2)){
			
			QualifyingProperties qp=new QualifyingProperties(sig);
			SignedProperties sp = new SignedProperties(sig, cert, claimedRoles, adr);
			sp.setDataObjectFormat(m_dataObject);
			qp.setSignedProperties(sp);
			sig.setQualifyingProperties(qp);
			Reference refsp = new Reference(si, sp);
			refsp.addTransformAlgorithm(CANONICALIZATION_METHOD_20010315);
			si.addReference(refsp);
		}
		sig.setSignedInfo(si);
		
		// add Signature
		addSignature(sig);
		return sig;
	}
	
	/**
	 * Adds a new Signature object
	 * @param attr Signature object to add
	 */
	public void addSignature(Signature sig) 
	{
		if(m_signatures == null)
			m_signatures = new ArrayList();
		m_signatures.add(sig);
	}
	
	/**
	 * return the desired Signature object
	 * @param idx index of the Signature object
	 * @return desired Signature object
	 */
	public Signature getSignature(int idx) {
		return (Signature)m_signatures.get(idx);
	}
	
	/**
	 * Removes the desired Signature object
	 * @param idx index of the Signature object
	 */
	public void removeSignature(int idx)
	{
		m_signatures.remove(idx);
	}
	
	/**
	 * return the latest Signature object
	 * @return desired Signature object
	 */
	public Signature getLastSignature() {
		return (Signature)m_signatures.get(m_signatures.size()-1);
	}
	
	/** 
	 * Deletes last signature
	 */
	public void removeLastSiganture()
	{
		if(m_signatures.size() > 0)
			m_signatures.remove(m_signatures.size()-1);
	}
	
	
	/**
	 * Helper method to create the xml header
	 * @return xml header
	 */
	private String xmlHeader()
	{
		StringBuffer sb = new StringBuffer("<?xml version=\"1.0\" encoding=\""+"UTF-8"+"\"?>\n");
		if (m_mode.equals(MODE_DETACHED_SIGNATURE))
			sb.append("<"+ROOT_NAME+">"); 
		return sb.toString();
	}
	
	/**
	 * Helper method to create the xml trailer
	 * @return xml trailer
	 */
	private String xmlTrailer()
	{
		if (m_mode.equals(MODE_DETACHED_SIGNATURE))
			return "</"+ROOT_NAME+">";
		else
			return "";
	}
	
	/**
	 * Converts the SignedDoc to XML form
	 * @return XML representation of SignedDoc
	 */
	public String toXML() throws DigiDocException 
	{
		StringBuffer sb = new StringBuffer(xmlHeader());
		
		if (m_mode.equals(MODE_ENVELOPED_SIGNATURE)){
			Builder builder=XOMBuilderFactory.getInstance(false);
			Document document;
			Content content=getContent();
			try {
				document=builder.build(new ByteArrayInputStream(content.getXML()));
				for(int i = 0; i < countSignatures(); i++) {
					Signature sig = getSignature(i);
					Document nodoSignature=builder.build(new ByteArrayInputStream(sig.getXML()));
					Node n=nodoSignature.getRootElement().copy();
					n.detach();
					document.getRootElement().appendChild(n);
				}
			} catch (DigiDocException e) {
				e.printStackTrace();
				throw e;
			} catch (Exception e) {
				e.printStackTrace();
				throw new DigiDocException(0,e.getMessage(),e);
			}
			sb.append(document.getRootElement().toXML());
		}else if (m_mode.equals(MODE_ENVELOPING_SIGNATURE)){
			for(int i = 0; i < countSignatures(); i++) {
				Signature sig = getSignature(i);
				String s = sig.toString();
				sb.append(s);
			}
		}else if(m_mode.equals(MODE_DETACHED_SIGNATURE)){
			Content df = getContent();
			String str = df.toString();
			sb.append(str);
			
			for(int i = 0; i < countSignatures(); i++) {
				Signature sig = getSignature(i);
				String s = sig.toString();
				sb.append(s);
			}
		}
		
		sb.append(xmlTrailer());        
		return sb.toString();
	}
	
	/**
	 * return the stringified form of SignedDoc
	 * @return SignedDoc string representation
	 */
	public String toString() 
	{
		String str = null;
		try {
			str = toXML();
		} catch(Exception ex) {}
		return str;
	}   
	
	/**
	 * Computes an SHA1 digest
	 * @param data input data
	 * @return SHA1 digest
	 */
	public static byte[] digest(byte[] data)
	throws DigiDocException 
	{
		byte[] dig = null;
		try {
			MessageDigest sha = MessageDigest.getInstance("SHA-1");
			dig = sha.digest(data);
		} catch(Exception ex) {
			DigiDocException.handleException(ex, DigiDocException.ERR_CALCULATE_DIGEST);
		}
		return dig;
	}
	
	/**
	 * Helper method for comparing
	 * digest values
	 * @param dig1 first digest value
	 * @param dig2 second digest value
	 * @return true if they are equal
	 */
	public static boolean compareDigests(byte[] dig1, byte[] dig2)
	{
		boolean ok = (dig1 != null) && (dig2 != null) && 
		(dig1.length == dig2.length);
		for(int i = 0; ok && (i < dig1.length); i++)
			if(dig1[i] != dig2[i])
				ok = false;
		return ok;
	}
	
	/** 
	 * Converts a hex string to byte array
	 * @param hexString input data
	 * @return byte array
	 */
	public static byte[] hex2bin(String hexString)
	{
		ByteArrayOutputStream bos = new ByteArrayOutputStream();
		try {
			for(int i = 0; (hexString != null) && 
			(i < hexString.length()); i += 2) {
				String tmp = hexString.substring(i, i+2);  
				Integer x = new Integer(Integer.parseInt(tmp, 16));
				bos.write(x.byteValue());    			
			}
		} catch(Exception ex) {
			System.err.println("Error converting hex string: " + ex);
		}
		return bos.toByteArray();
	}
	
	/**
	 * Converts a byte array to hex string
	 * @param arr byte array input data
	 * @return hex string
	 */
	public static String bin2hex(byte[] arr)
	{
		StringBuffer sb = new StringBuffer();
		for(int i = 0; i < arr.length; i++) {
			String str = Integer.toHexString((int)arr[i]);
			if(str.length() == 2)
				sb.append(str);
			if(str.length() < 2) {
				sb.append("0");
				sb.append(str);
			}
			if(str.length() > 2)
				sb.append(str.substring(str.length()-2));
		}
		return sb.toString();
	}
	
	private static byte[] parseContent(byte[] input) throws DigiDocException{
		byte[] result;
		Builder builder=XOMBuilderFactory.getInstance(false);
		
		try {
			result=Base64.decodeBytes(input);
		} catch (Exception e1) {
			//BASE64 Exception, maybe XML content?			
			try {
				builder.build(new ByteArrayInputStream(input));
			} catch (Exception e2) {
				return input;				
			}
			return input;
		}
		//Could it be an Base64 encoded XML?
		try {
			builder.build(new ByteArrayInputStream(result));
		} catch(Exception e){
			return input;
		}
		return result;
	}
	
	public String getFormatNamespace(){
		if (m_format.equals(FORMAT_XAdES_1_1_1))
			return XMLNS_ETSI_1_1_1;
//		else if (m_format.equals(FORMAT_XAdES_1_2_2))
//			return XMLNS_ETSI_1_2_2;
		else if (m_format.equals(FORMAT_XAdES_1_3_2))
			return XMLNS_ETSI_1_3_2;
		else
			return DEFAULT_XMLNS_ETSI;
	}
	
	private void createDocument() throws DigiDocException{
		if (m_mode.equals(MODE_ENVELOPED_SIGNATURE)){
			Content content=getContent();
			Element cnt=content.createElement();
			if (cnt.getParent()!=null)
				cnt.detach();
			m_xml=new Document(cnt);
			try {
				for(int i = 0; i < countSignatures(); i++) {
					Signature sig = getSignature(i);
					m_xml.getRootElement().appendChild(sig.createElement());
				}
			} catch (DigiDocException e) {
				e.printStackTrace();
				throw e;
			} catch (Exception e) {
				e.printStackTrace();
				throw new DigiDocException(0,e.getMessage(),e);
			}
		}else if (m_mode.equals(MODE_ENVELOPING_SIGNATURE)){
			if (countSignatures()>1){
				m_xml=new Document(new Element(SignedDoc.ROOT_NAME));
				for(int i = 0; i < countSignatures(); i++) {
					Signature sig = getSignature(i);
					m_xml.getRootElement().appendChild(sig.createElement());
				}		
			}else{
				m_xml=new Document(getSignature(0).createElement());
			}
		}else if(m_mode.equals(MODE_DETACHED_SIGNATURE)){
			m_xml=new Document(new Element(SignedDoc.ROOT_NAME));
			Content df = getContent();
			m_xml.getRootElement().appendChild(df.createElement());
			
			for(int i = 0; i < countSignatures(); i++) {
				Signature sig = getSignature(i);
				m_xml.getRootElement().appendChild(sig.createElement());
			}
		}
	}
	
	public Document getXML(){
		try {
			createDocument();
		} catch (DigiDocException e) {
			e.printStackTrace();
			return null;
		}
		return m_xml;
	}
	
	public static String guessFormat(byte[] signature) {
		String sig=new String(signature);
		if (sig.indexOf(XMLNS_ETSI_1_3_2)>0)
			return FORMAT_XAdES_1_3_2;
		else if (sig.indexOf(XMLNS_ETSI_1_1_1)>0)
			return FORMAT_XAdES_1_1_1; 
		else 
			return FORMAT_XMLDSIG;
	}
		
	private static Document m_counterSignatureObject=null;
	private static Element m_counterSignatureTemp=null;
	private static String m_counterSignatureId=null;
	
	public static byte[] prepareCounterSign(byte[] eSignature, String idToCounterSign, X509Certificate cert) throws DigiDocException{
		Builder builder=XOMBuilderFactory.getInstance(false);
		byte[] res=null;
		Document xml=null;
		try {
			xml=builder.build(new ByteArrayInputStream(eSignature));
		} catch (Exception e1) {
				throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"Error building XML document from data.",null);		
		}
		
		
		String format=guessFormat(eSignature);
		if (format.equals(FORMAT_XAdES_1_1_1)||
				format.equals(FORMAT_XAdES_1_2_2)||
				format.equals(FORMAT_XAdES_1_3_2)){
			
			XPathContext context=XPathContext.makeNamespaceContext(xml.getRootElement());
			//Obtenemos los prefijos de todo el documento
			HashMap namespaces=new HashMap();
			getNamespacesFromElementScope(xml.getRootElement(), namespaces);
			
			String prefixDS=(String) namespaces.get(XMLNS_XMLDSIG);
			String prefixXADES="";
			String namespaceXAdES=null;
			if (format.equals(FORMAT_XAdES_1_1_1)) {
				namespaceXAdES=XMLNS_ETSI_1_1_1;
				prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_1_1);
			}else if(format.equals(FORMAT_XAdES_1_2_2)) {
				namespaceXAdES=XMLNS_ETSI_1_2_2;
				prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_2_2);
			}else if(format.equals(FORMAT_XAdES_1_3_2)) {
				namespaceXAdES=XMLNS_ETSI_1_3_2;
				prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_3_2);
			}

			if(prefixDS == null)
				prefixDS = "ds";
			if(prefixXADES == null)
				prefixXADES = "xades";
			
			context.addNamespace(prefixXADES, namespaceXAdES);		
			context.addNamespace(prefixDS, XMLNS_XMLDSIG);
			
			if (prefixDS!= null && prefixDS.length() > 0)
				prefixDS=prefixDS+":";
			else
				prefixDS="ds:";
			
			if (prefixXADES!= null && prefixXADES.length() > 0)
				prefixXADES=prefixXADES+":";
			else
				prefixDS="xades:";						
			
			//Obtenemos todos los signature
			Nodes signatureNodes=xml.query(".//"+prefixDS+"Signature",context);
			
			//Modos de firma
			String[] mode=new String[signatureNodes.size()];
			for (int i=0;i<mode.length;i++) {
				mode[i]=guessSignatureMode((Element)signatureNodes.get(i));
			}
			
			if (signatureNodes.size()>0){
				Element signature=null;
				//SELECCIN DE SIGNATURE A CONTRAFIRMAR
				if (idToCounterSign!=null){
					for (int i=0;i<signatureNodes.size();i++){
						signature=(Element)signatureNodes.get(i);
						Attribute idAtt=signature.getAttribute("Id");
						if (idAtt!=null&&idAtt.getValue().equals(idToCounterSign)){
							break;
						}
					}
				}else{
					signature=(Element)signatureNodes.get(0);
				}
				if (signature!=null){
					
					Elements signValues=signature.getChildElements("SignatureValue",XMLNS_XMLDSIG);
					Element signValueTarget=null;
					Element unsignedProp=null;
					Element unsignedSignProp=null;
					
					if (signValues.size()==0){//must be a signatureValue below...
						throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"No SignatureValues found in document.",null);
					}
					
					signValueTarget=signValues.get(0);
					String xPathQuery;
					if (idToCounterSign!=null)
						xPathQuery="./"+prefixDS+"Object/"+prefixXADES+"QualifyingProperties[@Target=\"#"+idToCounterSign+"\"]";
					else {
						String idSig=signature.getAttributeValue("Id");
						if (!"".equals(idSig))
							xPathQuery="./"+prefixDS+"Object/"+prefixXADES+"QualifyingProperties[@Target=\"#"+idSig+"\"]";
						else
							xPathQuery="./"+prefixDS+"Object/"+prefixXADES+"QualifyingProperties";
					}
					Nodes qProperties= signature.query(xPathQuery,context);
					if (qProperties.size()==0){ //Posiblemente no tenga target
						xPathQuery="./"+prefixDS+"Object/"+prefixXADES+"QualifyingProperties";
						qProperties= signature.query(xPathQuery,context);
					}
					if (qProperties.size()>0) {//No debera tener ms de uno, por lo que cogemos el primero
						Element qp=(Element) qProperties.get(0);
						Elements unsigned=qp.getChildElements("UnsignedProperties",namespaceXAdES);
						if (unsigned.size()>0) {//Existe y marcamos el unsignedProperties
							unsignedProp=unsigned.get(0);
							unsigned = unsignedProp.getChildElements("UnsignedSignatureProperties",namespaceXAdES);
							if(unsigned.size() > 0){//Existe y marcamos el unsignedSignatureProperties
								unsignedSignProp = unsigned.get(0);
							}
							else{//No Existe el unsignedSignatureProperties, lo creamos y lo aadimos a UnsignedProperties
								unsignedSignProp=new Element("UnsignedSignatureProperties",namespaceXAdES);
								unsignedSignProp.setNamespaceURI(namespaceXAdES);
								unsignedSignProp.setNamespacePrefix(prefixXADES);
								unsignedProp.appendChild(unsignedSignProp);
							}
						}else {//No existe el UnsignedProperties; creamos el UnsignedProperties y su hijo UnsignedSignatureProperties
							unsignedProp=new Element("UnsignedProperties",namespaceXAdES);
							unsignedSignProp=new Element("UnsignedSignatureProperties",namespaceXAdES);
							unsignedSignProp.setNamespacePrefix(prefixXADES.replaceAll(":",""));
							unsignedProp.setNamespacePrefix(prefixXADES.replaceAll(":",""));
							unsignedProp.appendChild(unsignedSignProp);
							unsignedProp.setNamespaceURI(namespaceXAdES);
							unsignedSignProp.setNamespaceURI(namespaceXAdES);
							qp.appendChild(unsignedProp);
						}
					}else { //Es posible que exista un QualifyingPropertiesReference que apunte a otro sitio...
						xPathQuery="./"+prefixDS+"Object/"+prefixXADES+"QualifyingPropertiesReference";
						Nodes qPropertiesReferences= signature.query(xPathQuery,context);
						if (qPropertiesReferences.size()>0) {
							for (int i=0;i<qPropertiesReferences.size() && unsignedProp==null;i++) {//Buscamos alguno que contenga unsigned
								Element propRef=(Element)qPropertiesReferences.get(i);
								Attribute uri=propRef.getAttribute("URI");
								if (uri.getValue().startsWith("#")) {
									xPathQuery=".//*[@Id=\""+uri.getValue().substring(1)+"\"]";
									Nodes referencedProperties=xml.query(xPathQuery, context);
									if (referencedProperties.size()>0) {
										Elements unsigned=((Element)referencedProperties.get(0)).getChildElements("UnsignedProperties",namespaceXAdES);
										if (unsigned.size()>0) {//Existe y marcamos el unsignedProperties							
											unsignedProp=unsigned.get(0);
											unsigned = unsignedProp.getChildElements("UnsignedSignatureProperties",namespaceXAdES);
											if(unsigned.size() > 0){//Existe y marcamos el unsignedSignatureProperties
												unsignedSignProp = unsigned.get(0);
											}
										}
									}
								} else {
									throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"Cannot access to referenced URI: "+uri,null);
								}
							}
						}
					}
					
					//ultimo caso, muy raro
					if (unsignedProp==null) {//no se ha creado el unsignedProperties porque no se ha encontrado QualifyingProperties ni QualifyingPropertiesReference
						if(prefixXADES.indexOf(":") > 0)
							prefixXADES = prefixXADES.replaceAll(":","");
						if(prefixDS.indexOf(":") > 0)
							prefixDS = prefixDS.replaceAll(":","");
						Element object=new Element("Object",XMLNS_XMLDSIG);
						Element qp = new Element("QualifyingProperties",namespaceXAdES);						
						qp.setNamespacePrefix(prefixXADES);
						if (idToCounterSign!=null){ //Se crean y se aaden los atributos respecto al Id de la firma							
							Attribute qpattrId = new Attribute("Id",idToCounterSign+"-QualifyingProperties");
							Attribute qpattrTraget = new Attribute("Target","#"+idToCounterSign);
							qp.addAttribute(qpattrId);
							qp.addAttribute(qpattrTraget);
						}
						object.setNamespacePrefix(prefixDS);
						object.setNamespaceURI(XMLNS_XMLDSIG);
						signature.appendChild(object);
						unsignedProp=new Element("UnsignedProperties",namespaceXAdES);
						unsignedSignProp=new Element("UnsignedSignatureProperties",namespaceXAdES);
						unsignedProp.setNamespacePrefix(prefixXADES);
						unsignedSignProp.setNamespacePrefix(prefixXADES);
						unsignedProp.setNamespaceURI(namespaceXAdES);
						unsignedSignProp.setNamespaceURI(namespaceXAdES);
						unsignedProp.appendChild(unsignedSignProp);
						qp.appendChild(unsignedProp);
						object.appendChild(qp);
					}
					
					CounterSignature cs=new CounterSignature(signValueTarget,cert,new SignedDoc(format,VERSION_1_0,MODE_DETACHED_SIGNATURE));
					m_counterSignatureTemp=cs.getElement();
					unsignedSignProp.appendChild(m_counterSignatureTemp);
					res=cs.getSignatureMember().getSignedInfo().toCanonicalXML();
					m_counterSignatureId=cs.getSignatureValueId();
				}
			}else{
				throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"No signatures found in document.",null);
			}
			
		}else if (format.equals(FORMAT_XMLDSIG)) {
			throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"CounterSignatures are disallowed in XMLDSig format. May use XAdES instead",null);
		}
		m_counterSignatureObject=xml;
		return res;
	}
	
	public static byte[] finishCounterSignature(byte[] counterSignatureValue) throws DigiDocException {
		try {
			if (m_counterSignatureId!=null &&
					m_counterSignatureObject!=null &&
					m_counterSignatureTemp!=null) {
				
				XPathContext context=XPathContext.makeNamespaceContext(m_counterSignatureTemp);
				
				HashMap namespaces=new HashMap();
				getNamespacesFromElementScope(m_counterSignatureTemp, namespaces);
				
				String prefixDS=(String) namespaces.get(XMLNS_XMLDSIG);
				context.addNamespace(prefixDS, XMLNS_XMLDSIG);
				
				String xPathQuery=".//*[@Id=\""+m_counterSignatureId+"\"]";
				System.out.println("Searching for: "+m_counterSignatureId);
				Nodes n=m_counterSignatureTemp.query(xPathQuery,context);
				if (n.size()!=1) {
					throw new DigiDocException(0,"ERROR",null);
				}
				Element svElement=(Element)n.get(0);
				svElement.removeChildren();
				String base64SignatureValue=Base64.encodeBytes(counterSignatureValue);
				svElement.appendChild(base64SignatureValue);
				return m_counterSignatureObject.toXML().getBytes();
			}else
				throw new DigiDocException(DigiDocException.ERR_ELEMENT_INIT,"CounterSignature not started",null);
		}catch(DigiDocException e) {
			throw e;
		}catch(Exception e) {
			throw new DigiDocException(0,"XXXXX",e);
		}finally {
			m_counterSignatureId=null;
			m_counterSignatureObject=null;
			m_counterSignatureTemp=null;
		}
	}
		
	public static byte[] prepareCoSign(byte[] eSignature, byte[] data, String idToCoSign, X509Certificate cert, String encoding, String algHash) throws DigiDocException, UnsupportedEncodingException{
		Builder builder=XOMBuilderFactory.getInstance(false);
		byte[] res=null;
		byte[] originalData = null;
		Document xml=null;
		Document originalDataXML=null;
		Element recoveredData=null;
		if (eSignature==null||data==null||cert==null)
			throw new NullPointerException("Invalid parameter");
				
		try {
			xml=builder.build(new ByteArrayInputStream(eSignature));
		} catch (Exception e) {
			throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"Error building XML document from data.",null);
		}
		
		try {
			originalDataXML=builder.build(new ByteArrayInputStream(data));
		} catch (Exception e) {
			// No son datos XML
			originalData=parseContent(data);			
		}

		
		String format=guessFormat(eSignature);
		
		XPathContext context=XPathContext.makeNamespaceContext(xml.getRootElement());
		
		//Obtenemos los prefijos de todo el documento
		HashMap namespaces=new HashMap();
		getNamespacesFromElementScope(xml.getRootElement(), namespaces);
		
		String prefixDS=(String) namespaces.get(XMLNS_XMLDSIG);
		String prefixXADES="";
		String namespaceXAdES=null;
		
		if (format.equals(FORMAT_XAdES_1_1_1)) {
			namespaceXAdES=XMLNS_ETSI_1_1_1;
			prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_1_1);
		}else if(format.equals(FORMAT_XAdES_1_2_2)) {
			namespaceXAdES=XMLNS_ETSI_1_2_2;
			prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_2_2);
		}else if(format.equals(FORMAT_XAdES_1_3_2)) {
			namespaceXAdES=XMLNS_ETSI_1_3_2;
			prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_3_2);
		}
		
//		else if(format.equals(FORMAT_XMLDSIG)) {
//			//La cofirma tendra formato XAdES
//			namespaceXAdES=XMLNS_ETSI_1_3_2;
//			prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_3_2);
//		}
		
		if(prefixDS == null)
			prefixDS = "ds";
		if(prefixXADES == null)
			prefixXADES = "xades";
		
		context.addNamespace(prefixXADES, namespaceXAdES);		
		context.addNamespace(prefixDS, XMLNS_XMLDSIG);
		
		if (prefixDS!= null && prefixDS.length() > 0)
			prefixDS=prefixDS+":";
		else
			prefixDS="ds:";
		
		if (prefixXADES!= null && prefixXADES.length() > 0)
			prefixXADES=prefixXADES+":";
		else
			prefixDS="xades:";
		
		//Obtenemos todos los signature		
		Nodes signatureNodes=xml.query("//"+prefixDS+"Signature",context);
		
		if (signatureNodes.size()>0){
			Element signature=null;
			if (idToCoSign!=null){
				for (int i=0;i<signatureNodes.size();i++){
					signature=(Element)signatureNodes.get(i);
					Attribute idAtt=signature.getAttribute("Id");
					if (idAtt!=null&&idAtt.getValue().equals(idToCoSign)){
						break;
					}
				}
			}else{
				//Cogemos por defecto la primera que encontramos
				signature=(Element)signatureNodes.get(0);
			}
			if (signature!=null){
				String mode=guessSignatureMode(signature);
				if (MODE_ENVELOPED_SIGNATURE.equals(mode)) {
					//Obtenemos el padre
					recoveredData=(Element)xml.getRootElement().copy();					
					String xPathQuery="./"+prefixDS+"Signature";
					Nodes signatures=recoveredData.query(xPathQuery, context);
					
					//Quitamos las firmas para dejar solo los datos					
					for (int i=0;i<signatures.size();i++) {
						Node signToRemove=signatures.get(i);
						recoveredData.removeChild(signToRemove);
					}										
					
					//Para comprobar que son iguales vamos a canonicalizar
					CanonicalizationFactory fact=SignedDoc.getCanonicalizeInstance();
					byte[] comp1=fact.canonicalize(recoveredData, Canonicalizer.CANONICAL_XML);
					byte[] comp2=null;
					if(originalDataXML != null)
						comp2=fact.canonicalize(originalDataXML.getRootElement(), Canonicalizer.CANONICAL_XML);
					else if(originalData != null)
						comp2=originalData;
					else
						throw new DigiDocException(DigiDocException.ERR_DATA_FILE_CONTENT_TYPE,"Information to cosign lost.",null);
					if(!Arrays.equals(comp1, comp2))
						throw new DigiDocException(0,"Original data not the same that reovered data",null);
					SignedDoc actualDoc=new SignedDoc(format,VERSION_1_0,MODE_ENVELOPED_SIGNATURE);
					actualDoc.setContentEnveloped(data, null,"application/xml",encoding);
					
					m_cosign=actualDoc.prepareSignature(cert, null, null);
					m_coSignatureObject=xml;
					m_coSignatureTemp=signature;
					return m_cosign.getSignedInfo().toCanonicalXML();
				}else if(MODE_ENVELOPING_SIGNATURE.equals(mode)) {
					//Caso enveloping
					Element root=null;
					if (signature.getParent()==null||signature.getDocument().getRootElement().equals(signature)) {
						root=new Element(ROOT_NAME);
						signature.getDocument().setRootElement(new Element("NUL"));
						signature.detach();
						root.appendChild(signature);
					}else {
						root=(Element) signature.getParent();
						root.getDocument().setRootElement(new Element("NUL"));
					}
					//Comparar datos
					//No es necesario, ya que la cofirma enveloping se hace explcita
					
					//Firmamos
					SignedDoc actualDoc=new SignedDoc(format,VERSION_1_0,MODE_ENVELOPING_SIGNATURE);
					actualDoc.setContent(data,null,null,null,algHash);
					m_cosign=actualDoc.prepareSignature(cert, null, null);
					m_coSignatureObject=new Document(root);
					m_coSignatureTemp=signature;
					
					return m_cosign.getSignedInfo().toCanonicalXML();
				}else if(MODE_DETACHED_SIGNATURE.equals(mode)) {
					
					SignedDoc actualDoc=new SignedDoc(format,VERSION_1_0,MODE_DETACHED_SIGNATURE);
					
					String idToData=guessDetachedDataURI(signature,prefixDS,context);
					if (idToData!=null) {
						//Intentamos acceder a el desde el rootElement
						String xPathQuery=".//*[@Id=\""+idToData.replaceFirst("#","")+"\"]";
						Nodes elementWithId=xml.query(xPathQuery, context);						
						if (elementWithId.size()!=1) {
							throw new DigiDocException(0,"Invalid number of elements with id "+idToData,null);
						}				
						Content content=Content.parseContent((Element)elementWithId.get(0),actualDoc);
						actualDoc.m_content=content;						
						m_cosign=actualDoc.prepareSignature(cert, null, null);
						m_coSignatureObject=xml;
						m_coSignatureTemp=signature;
						return m_cosign.getSignedInfo().toCanonicalXML();
					}
				}else {
					throw new DigiDocException(0,"INVALID SIGNATURE MODE",null);
				}
				
			}
		}else{
			throw new DigiDocException(DigiDocException.ERR_PARSE_XML,"No signatures found in document.",null);
		}		
		
		return res;
	}
	
	public static byte[] finishCoSignature(byte[] coSignatureValue) throws DigiDocException {
		try {
			if (m_cosign!=null &&
					m_coSignatureObject!=null &&
					m_coSignatureTemp!=null) {
				m_cosign.setSignatureValue(coSignatureValue);
				Element newSignature=m_cosign.getElement();
				newSignature.detach();
				ParentNode pn=m_coSignatureTemp.getParent();
				pn.appendChild(newSignature);
				return m_coSignatureObject.toXML().getBytes();
			}else
				throw new DigiDocException(DigiDocException.ERR_ELEMENT_INIT,"CounterSignature not started",null);
		}catch(DigiDocException e) {
			throw e;
		}catch(Exception e) {
			throw new DigiDocException(0,"XXXXX",e);
		}finally {
			m_cosign=null;
			m_coSignatureObject=null;
			m_counterSignatureTemp=null;
		}
	}
	
	
	static void getNamespacesFromElementScope(Element e,Map namespaces) {
		Elements children=e.getChildElements();
		
		//Insertamos el prefix y namespace actual
		String prefix=e.getNamespacePrefix();
		String namespace=e.getNamespaceURI();
		if (prefix!=null && !prefix.equals("") && !namespaces.containsKey(prefix)) {
			namespaces.put(namespace,prefix);
		}
		
		//Comprobamos si hay mas
		for (int i=1;i<e.getNamespaceDeclarationCount();i++) {
			String additionalPrefix=e.getNamespacePrefix(i);
			String additionalNamespace=e.getNamespaceURI(additionalPrefix);
			if (additionalPrefix!=null && !additionalPrefix.equals("") && !namespaces.containsKey(additionalPrefix)) {
				namespaces.put(additionalNamespace,additionalPrefix);
			} 
		}
		
		for (int i=0;i<children.size();i++) {
			Element child=children.get(i);
			getNamespacesFromElementScope(child, namespaces);
		}
	}
	
	static String guessDetachedDataURI(Element signature, String prefixDS, XPathContext context) {
		Nodes references=signature.query("./"+prefixDS+"SignedInfo/"+prefixDS+"Reference",context);
		
		for(int i=references.size()-1;i>=0;i--) {
			Element reference=(Element)references.get(i);
			Attribute uri=reference.getAttribute("URI");
			String idUri=uri.getValue();
			if (idUri.startsWith("#")) {//posibilidad de enveloping o detached afirma
				Nodes elementWithId=signature.query(".//*[@Id=\""+idUri.substring(1)+"\"]", context);
				if (elementWithId.size()==0) {//Hacemos referencia a un id externo. Detached
					return idUri;
				}
			}
		}
		return null;
	}
	
	static String guessSignatureMode(Element xml) {
		//Obtenemos el contexto de la firma
		XPathContext context=XPathContext.makeNamespaceContext(xml);
		
		HashMap namespaces=new HashMap();
		getNamespacesFromElementScope(xml,namespaces);
		
		String prefixDS=(String) namespaces.get(XMLNS_XMLDSIG);
		String prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_1_1);
		if (prefixXADES==null)
			prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_2_2);
		if (prefixXADES==null)
			prefixXADES=(String) namespaces.get(XMLNS_ETSI_1_3_2);
		Nodes references;
		if (!"".equals(prefixDS)) {
			prefixDS=prefixDS+":";
		}else {
			prefixDS="";
		}
		references=xml.query("./"+prefixDS+"SignedInfo/"+prefixDS+"Reference",context);
		
		//Partimos de que si tenemos algn reference
		for(int i=references.size()-1;i>=0;i--) {
			Element reference=(Element)references.get(i);
			Attribute uri=reference.getAttribute("URI");
			if (uri==null||uri.getValue().equals("")) {//Posible enveloped
				Nodes envelopedTransforms=reference.query("./"+prefixDS+"Transforms/"+prefixDS+"Transform[@Algorithm=\""+SignedDoc.ENVELOPED_TRANSFORM+"\"]", context);
				if (envelopedTransforms.size()>0){
					//Es una firma enveloped				
					return MODE_ENVELOPED_SIGNATURE;
				}
				else {
					envelopedTransforms=reference.query("./"+prefixDS+"Transforms/"+prefixDS+"Transform[@Algorithm=\""+TransformXPath.implementedTransformURI+"\"]", context);
					if (envelopedTransforms.size()>0){
						//Es una firma enveloped	
						return MODE_ENVELOPED_SIGNATURE;
					}
					else{
						//Fallo en la firma
						return null;
					}
				}
			}else {
				String idUri=uri.getValue();
				if (idUri.startsWith("#")) {//posibilidad de enveloping o detached afirma
					Nodes elementWithId=xml.query(".//*[@Id=\""+idUri.substring(1)+"\"]", context);
					if (elementWithId.size()==0) {//Hacemos referencia a un id externo. Detached
						return MODE_DETACHED_SIGNATURE;
					}
				}
			}
		}
		//Todas las reference analizadas y no es ni detached ni enveloped, va a ser que es enveloping
		return MODE_ENVELOPING_SIGNATURE;
	}
	
	public static byte[] encodingBytes(byte[] in, String encoding){		
        ByteArrayOutputStream baos=new ByteArrayOutputStream();
        ByteArrayInputStream bais = new ByteArrayInputStream(in);
        Builder builder = new Builder();
        if (encoding == null)
        	return in;
        try {
        	Document doc = builder.build(bais);
	        Serializer serializer=new Serializer(baos,encoding);
	        serializer.write(doc);
		} catch (UnsupportedEncodingException e) {
			System.out.println("Unsupported encoding '"+encoding+"': "+e.getCause());
			e.printStackTrace();
		} catch (IOException e) {
			System.out.println("Error writing document: "+e.getCause());
			e.printStackTrace();
		} catch (ValidityException e) {
			System.out.println("Error validating XML document: "+e.getCause());
			e.printStackTrace();
		} catch (ParsingException e) {
			System.out.println("Error parsing XML document: "+e.getCause());
			e.printStackTrace();
		}
        return baos.toByteArray();
	}
	
	public byte[] getContentEnveloped() throws DigiDocException{
		byte[] contentEnveloped = null;
		try {
			Nodes ns = null;
			Element contentElem = null;
			contentElem = m_content.getElement();
	        XPathContext context=XPathContext.makeNamespaceContext(contentElem);
	        context.addNamespace("xades", XMLNS_ETSI_1_3_2);      
	        context.addNamespace("ds", XMLNS_XMLDSIG);
			ns = contentElem.query("//ds:Signature",context);
			for (int i=0 ; i<ns.size(); i++){
				try{
					contentElem.removeChild(ns.get(i));
				} catch (NoSuchChildException e){

				} catch (Exception e){
					System.out.println("Error: " + e.getCause());
				}
			}
			//Canonicalizamos el contenido a firmar
			CanonicalizationFactory canonFact=SignedDoc.getCanonicalizeInstance();
			contentEnveloped = canonFact.canonicalize(contentElem, Canonicalizer.CANONICAL_XML);
		} catch (DigiDocException e) {
			throw e;
		}
		
		return contentEnveloped;
	}
}
