/*
 * SignedProperties.java
 * PROJECT: JDigiDoc
 * DESCRIPTION: Digi Doc functions for creating
 *	and reading signed documents. 
 * AUTHOR:  Veiko Sinivee, S|E|B IT Partner Estonia
 *==================================================
 * Copyright (C) AS Sertifitseerimiskeskus
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * GNU Lesser General Public Licence is available at
 * http://www.gnu.org/copyleft/lesser.html
 *==================================================
 *
 * THIS CLASS HAS BEEN MODIFIED:
 * -----------------------------
 * MODIFICATION: Output has been modified to not insert 'newline' characters.
 * AUTHOR: Ministerio de Administraciones Pblicas. Gobierno de Espaa.
 */

package ee.sk.digidoc;
import java.io.IOException;
import java.io.Serializable;
import java.math.BigInteger;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.spi.CharsetProvider;
import java.security.Principal;
import java.security.Provider;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Date;

import javax.security.auth.x500.X500Principal;

import sun.security.x509.X500Name;

import nu.xom.Attribute;
import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import ee.sk.digidoc.factory.CanonicalizationFactory;
import ee.sk.digidoc.factory.XOMBuilderFactory;
import ee.sk.digidoc.utils.ConvertUtils;
import ee.sk.digidoc.utils.XMLObject;



/**
 * Models the SignedProperties element of
 * an XML-DSIG/ETSI Signature.
 * @author  Veiko Sinivee
 * @version 1.0
 */
public class SignedProperties extends XMLObject implements Serializable
{
	private static final String LOCALNAME = "SignedProperties";

	/** signature object to which this belongs */
	private QualifyingProperties m_qualifyProp;
	
	private Signature m_sig;
	/** id attribute */
	private String m_id;
	/** target attribute */
	private String m_target;
	/** signing time measured by signers own computer */
	private Date m_signingTime;
	/** signers certs digest algorithm */
	private String m_certDigestAlgorithm;
	/** signers cert id */
	private String m_certId;
	/** signers certs digest data */
	private byte[] m_certDigestValue;
	/** signers certs issuer serial number */
	private BigInteger m_certSerial;
	/** digest over the original bytes read from XML file  */
	private byte[] m_origDigest;
	
	/** signature production place */
	private SignatureProductionPlace m_address;
	/** claimed roles */
	private ArrayList m_claimedRoles;
	
	private SignaturePolicyIdentitifier m_policy;
	
	private DataObjectFormat m_dataObjectFormat;
	
	/** 
	 * Creates new SignedProperties. Initializes
	 * everything to null
	 * @param sig parent signature
	 */
	public SignedProperties(Signature sig) {
		m_sig = sig;
		m_qualifyProp=sig.getQualifyingProperties();
		m_id = null;
		m_target = null;
		m_signingTime = null;
		m_certDigestAlgorithm = null;
		m_certDigestValue = null;
		m_certSerial = null;
		m_claimedRoles = null;
		m_address = null;
		m_certId = null;
		m_origDigest = null;
		m_dataObjectFormat=null;
	}
	
	/** 
	 * Creates new SignedProperties. 
	 * @param sig parent signature
	 * @param id id attribute value
	 * @param target target attribute value
	 * @param signingTime signing timestamp
	 * @param certId signers cert id (in XML)
	 * @param certDigAlg signers cert digest algorithm id/uri
	 * @param digest signers cert digest value
	 * @param serial signers cert serial number
	 * @throws DigiDocException for validation errors
	 */
	public SignedProperties(Signature sig, String id, String target, Date signingTime, 
			String certId, String certDigAlg, byte[] digest, BigInteger serial) 
	throws DigiDocException
	{
		m_sig = sig;
		m_qualifyProp=sig.getQualifyingProperties();
		if (id==null)
			throw new DigiDocException(DigiDocException.ERR_SIGPROP_ID,"Id must be in form: <signature-id>-SignedProperties", null);
		m_id=id;
		if(target == null)
			throw new DigiDocException(DigiDocException.ERR_SIGPROP_TARGET,"Target must be in form: #<signature-id>", null);
		m_target=target;
		if(signingTime == null)
			throw new DigiDocException(DigiDocException.ERR_SIGNING_TIME, "Signing time cannot be empty!", null);
		m_signingTime=signingTime;
		if(certId == null)
			throw new DigiDocException(DigiDocException.ERR_SIGPROP_CERT_ID, 
					"Cert Id must be in form: <signature-id>-CERTINFO", null);
		m_certId=certId;
		if(certDigAlg == null || !certDigAlg.equals(SignedDoc.SHA1_DIGEST_ALGORITHM))
			throw new DigiDocException(DigiDocException.ERR_CERT_DIGEST_ALGORITHM, 
					"Currently supports only SHA1 digest algorithm", null);
		m_certDigestAlgorithm=certDigAlg;
		if(digest == null || digest.length != SignedDoc.SHA1_DIGEST_LENGTH)
			throw new DigiDocException(DigiDocException.ERR_DIGEST_LENGTH, 
					"SHA1 digest data is allways 20 bytes of length", null);
		m_certDigestValue=digest;
		
		setCertSerial(serial);
		m_claimedRoles = null;
		m_address = null;
		m_origDigest = null;
	}
	
	/** 
	 * Creates new SignedProperties with default
	 * values taken from signers certificate and signature
	 * @param sig Signature reference
	 * @param cert signers certificate
	 * @param claimedRoles signers claimed roles
	 * @param adr signers address
	 * @throws DigiDocException for validation errors
	 */
	public SignedProperties(Signature sig, X509Certificate cert,
			String[] claimedRoles, SignatureProductionPlace adr) 
	throws DigiDocException
	{
		m_sig = sig;
		m_id=sig.getId() + "-SignedProperties";
		m_target="#" + sig.getId();
		m_signingTime=new Date();
		m_certId=sig.getId() + "-CERTINFO";
		m_certDigestAlgorithm=SignedDoc.SHA1_DIGEST_ALGORITHM;
		try {
			m_certDigestValue=SignedDoc.digest(cert.getEncoded());
		} catch(Exception ex) {
			DigiDocException.handleException(ex, DigiDocException.ERR_CALCULATE_DIGEST);
		}
		m_certSerial=cert.getSerialNumber();
		if((claimedRoles != null) && (claimedRoles.length > 0)) {
			for(int i = 0; i < claimedRoles.length; i++) 
				addClaimedRole(claimedRoles[i]);
		}
		if(adr != null)
			m_address=adr;
		m_origDigest = null;
		createElement();
	}
	
	public SignedProperties(Signature sig, X509Certificate cert,
			String[] claimedRoles, SignatureProductionPlace adr, SignaturePolicyIdentitifier spi) 
	throws DigiDocException
	{
		m_sig = sig;
		m_id=sig.getId() + "-SignedProperties";
		m_target="#" + sig.getId();
		m_signingTime=new Date();
		m_certId=sig.getId() + "-CERTINFO";
		m_certDigestAlgorithm=SignedDoc.SHA1_DIGEST_ALGORITHM;
		try {
			m_certDigestValue=SignedDoc.digest(cert.getEncoded());
		} catch(Exception ex) {
			DigiDocException.handleException(ex, DigiDocException.ERR_CALCULATE_DIGEST);
		}
		m_certSerial=cert.getSerialNumber();
		if((claimedRoles != null) && (claimedRoles.length > 0)) {
			for(int i = 0; i < claimedRoles.length; i++) 
				addClaimedRole(claimedRoles[i]);
		}
		m_address=adr;
		m_policy=spi;
		m_origDigest = null;
		createElement();
	}
	
	/**
	 * Accessor for id attribute
	 * @return value of id attribute
	 */
	public String getId() {
		return m_id;
	}
	
	/**
	 * Mutator for id attribute
	 * @param str new value for id attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setId(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateId(str);
		if(ex != null)
			throw ex;
		m_id = str;
		createElement();
	}
	
	/**
	 * Accessor for origDigest attribute
	 * @return value of origDigest attribute
	 */
	public byte[] getOrigDigest() {
		return m_origDigest;
	}
	
	/**
	 * Mutator for origDigest attribute
	 * @param str new value for origDigest attribute
	 * @throws DigiDocException 
	 */    
	public void setOrigDigest(byte[] data) throws DigiDocException 
	{
		m_origDigest = data;
		createElement();
	}
	
	/**
	 * Helper method to validate an id
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateId(String str)
	{
		DigiDocException ex = null;
		if(str == null)
			ex = new DigiDocException(DigiDocException.ERR_SIGPROP_ID, 
					"Id must be in form: <signature-id>-SignedProperties", null);
		return ex;
	}
	
	/**
	 * Accessor for target attribute
	 * @return value of target attribute
	 */
	public String getTarget() {
		return m_target;
	}
	
	/**
	 * Mutator for target attribute
	 * @param str new value for target attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setTarget(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateTarget(str);
		if(ex != null)
			throw ex;
		m_target = str;
		createElement();
	}
	
	/**
	 * Helper method to validate a target
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateTarget(String str)
	{
		DigiDocException ex = null;
		if(str == null)
			ex = new DigiDocException(DigiDocException.ERR_SIGPROP_TARGET, 
					"Target must be in form: #<signature-id>", null);
		return ex;
	}
	
	/**
	 * Accessor for certId attribute
	 * @return value of certId attribute
	 */
	public String getCertId() {
		return m_certId;
	}
	
	/**
	 * Mutator for certId attribute
	 * @param str new value for certId attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setCertId(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateCertId(str);
		if(ex != null)
			throw ex;
		m_certId = str;
		createElement();
	}
	
	/**
	 * Helper method to validate an certificate id
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateCertId(String str)
	{
		DigiDocException ex = null;
		if(str == null)
			ex = new DigiDocException(DigiDocException.ERR_SIGPROP_CERT_ID, 
					"Cert Id must be in form: <signature-id>-CERTINFO", null);
		return ex;
	}
	
	/**
	 * Accessor for signatureProductionPlace attribute
	 * @return value of signatureProductionPlace attribute
	 */
	public SignatureProductionPlace getSignatureProductionPlace() {
		return m_address;
	}
	
	/**
	 * Mutator for signatureProductionPlace attribute
	 * @param str new value for signatureProductionPlace attribute
	 */    
	public void setSignatureProductionPlace(SignatureProductionPlace adr) 
	throws DigiDocException
	{
		m_address = adr;
		createElement();
	}
	
	/**
	 * Accessor for signingTime attribute
	 * @return value of signingTime attribute
	 */
	public Date getSigningTime() {
		return m_signingTime;
	}
	
	/**
	 * Mutator for signingTime attribute
	 * @param str new value for signingTime attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setSigningTime(Date d) 
	throws DigiDocException
	{
		DigiDocException ex = validateSigningTime(d);
		if(ex != null)
			throw ex;
		m_signingTime = d;
		createElement();
	}
	
	/**
	 * Helper method to validate a signingTime
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateSigningTime(Date d)
	{
		DigiDocException ex = null;
		if(d == null) // check the uri somehow ???
			ex = new DigiDocException(DigiDocException.ERR_SIGNING_TIME, 
					"Singing time cannot be empty!", null);
		return ex;
	}
	
	/**
	 * Accessor for certDigestAlgorithm attribute
	 * @return value of certDigestAlgorithm attribute
	 */
	public String getCertDigestAlgorithm() {
		return m_certDigestAlgorithm;
	}
	
	/**
	 * Mutator for certDigestAlgorithm attribute
	 * @param str new value for certDigestAlgorithm attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setCertDigestAlgorithm(String str) 
	throws DigiDocException
	{
		DigiDocException ex = validateCertDigestAlgorithm(str);
		if(ex != null)
			throw ex;
		m_certDigestAlgorithm = str;
		createElement();
	}
	
	/**
	 * Helper method to validate a digest algorithm
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateCertDigestAlgorithm(String str)
	{
		DigiDocException ex = null;
		if(str == null || !str.equals(SignedDoc.SHA1_DIGEST_ALGORITHM))
			ex = new DigiDocException(DigiDocException.ERR_CERT_DIGEST_ALGORITHM, 
					"Currently supports only SHA1 digest algorithm", null);
		return ex;
	}
	
	/**
	 * Accessor for certDigestValue attribute
	 * @return value of certDigestValue attribute
	 */
	public byte[] getCertDigestValue() {
		return m_certDigestValue;
	}
	
	/**
	 * Mutator for certDigestValue attribute
	 * @param data new value for certDigestValue attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setCertDigestValue(byte[] data) 
	throws DigiDocException
	{
		DigiDocException ex = validateCertDigestValue(data);
		if(ex != null)
			throw ex;
		m_certDigestValue = data;
		createElement();
	}
	
	/**
	 * Helper method to validate a digest value
	 * @param data input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateCertDigestValue(byte[] data)
	{
		DigiDocException ex = null;
		if(data == null || data.length != SignedDoc.SHA1_DIGEST_LENGTH)
			ex = new DigiDocException(DigiDocException.ERR_DIGEST_LENGTH, 
					"SHA1 digest data is allways 20 bytes of length", null);
		return ex;
	}
	
	/**
	 * Accessor for certSerial attribute
	 * @return value of certSerial attribute
	 */
	public BigInteger getCertSerial() {
		return m_certSerial;
	}
	
	/**
	 * Mutator for certSerial attribute
	 * @param str new value for certSerial attribute
	 * @throws DigiDocException for validation errors
	 */    
	public void setCertSerial(BigInteger i) 
	throws DigiDocException
	{
		DigiDocException ex = validateCertSerial(i);
		if(ex != null)
			throw ex;
		m_certSerial = i;
		createElement();
	}
	
	/**
	 * Helper method to validate a certSerial
	 * @param str input data
	 * @return exception or null for ok
	 */
	private DigiDocException validateCertSerial(BigInteger i)
	{
		DigiDocException ex = null;
		if(i == null) // check the uri somehow ???
			ex = new DigiDocException(DigiDocException.ERR_CERT_SERIAL, 
					"Certificates serial number cannot be empty!", null);
		return ex;
	}
	
	/**
	 * Returns the count of claimedRole objects
	 * @return count of Reference objects
	 */
	public int countClaimedRoles() {
		return ((m_claimedRoles == null) ? 0 : m_claimedRoles.size());
	}
	
	/**
	 * Adds a new reference object
	 * @param ref Reference object to add
	 */
	public void addClaimedRole(String role) 
	{
		if(m_claimedRoles == null)
			m_claimedRoles = new ArrayList();
		m_claimedRoles.add(role);
	}
	
	/**
	 * Returns the desired claimedRole object
	 * @param idx index of the claimedRole object
	 * @return desired claimedRole object
	 */
	public String getClaimedRole(int idx) {
		return (String)m_claimedRoles.get(idx);
	}
		
	public SignaturePolicyIdentitifier getSignaturePolicy() {
		return m_policy;
	}

	
	public void setSignaturePolicy(SignaturePolicyIdentitifier policy) {
		this.m_policy = policy;
	}
	
	

	/**
	 * Helper method to validate the whole
	 * SignedProperties object
	 * @return a possibly empty list of DigiDocException objects
	 */
	public ArrayList validate()
	{
		ArrayList errs = new ArrayList();
		DigiDocException ex = validateId(m_id);
		if(ex != null)
			errs.add(ex);
		ex = validateTarget(m_target);
		if(ex != null)
			errs.add(ex);
		ex = validateCertId(m_certId);
		if(ex != null)
			errs.add(ex);
		ex = validateSigningTime(m_signingTime);
		if(ex != null)
			errs.add(ex);
		ex = validateCertDigestAlgorithm(m_certDigestAlgorithm);
		if(ex != null)
			errs.add(ex);
		ex = validateCertDigestValue(m_certDigestValue);
		if(ex != null)
			errs.add(ex);
		ex = validateCertSerial(m_certSerial);
		if(ex != null)
			errs.add(ex);  
		// claimed roles
		// and signature production place are optional
		return errs;
	}
	
	
	
	/**
	 * Calculates the digest of SignedProperties block
	 * @return SignedProperties block digest
	 */
	public byte[] calculateDigest()
	throws DigiDocException
	{
		if(m_origDigest == null) {
			CanonicalizationFactory canFac=SignedDoc.getCanonicalizeInstance();
			byte[] tmp = canFac.canonicalize(getElement(),SignedDoc.CANONICALIZATION_METHOD_20010315);
			return SignedDoc.digest(tmp);
		}
		else
			return m_origDigest;
	}
	
	public Element createElement(String namespace) throws DigiDocException{
		return createElement();
	}
	
	public Element createElement() throws DigiDocException{
		String format=m_sig.getSignedDoc().getFormat();
		String namespace=null;
		if (format.equals(SignedDoc.FORMAT_XAdES_1_1_1))
			namespace=SignedDoc.XMLNS_ETSI_1_1_1;
		else if (format.equals(SignedDoc.FORMAT_XAdES_1_2_2))
			namespace=SignedDoc.XMLNS_ETSI_1_2_2;
		else if (format.equals(SignedDoc.FORMAT_XAdES_1_3_2))
			namespace=SignedDoc.XMLNS_ETSI_1_3_2;
		if(thisObject==null)
			thisObject=makeEasyElement("SignedProperties",namespace);
		else if(thisObject!=null&&m_sig.isExternalSignature()){
			return thisObject;
		}
		if (m_id!=null)
			thisObject.addAttribute(new Attribute("Id",m_id));
		if (thisObject.getChildCount()>0){
			thisObject.removeChildren();
		}
		
		if (m_dataObjectFormat!=null) {
//			if (thisObject.getChildElements("SignedDataObjectProperties", namespace).size()>0) {
//				thisObject.removeChild()
//			}
			Element sdop=makeEasyElement("SignedDataObjectProperties",namespace);
			Element dataObjectFormat=m_dataObjectFormat.createElement(namespace);
			dataObjectFormat.detach();
			sdop.appendChild(dataObjectFormat);
			thisObject.appendChild(sdop);
		}else {
			
		}
		
		Element signedSignatureProps=makeEasyElement("SignedSignatureProperties",namespace);
		if (m_signingTime!=null){;
			Element signingTime=makeEasyElement("SigningTime",namespace);
			signingTime.appendChild(ConvertUtils.date2string(m_signingTime, m_sig.getSignedDoc()));
			signedSignatureProps.appendChild(signingTime);
		}
		
		
		
		if ((m_certDigestAlgorithm!=null&&m_certDigestValue!=null)||
				(m_sig.getKeyInfo()!=null&&m_certSerial!=null)){
			Element signingCertificate=makeEasyElement("SigningCertificate",namespace);
			Element cert=makeEasyElement("Cert",namespace);	
			if (m_certDigestAlgorithm!=null&&
					m_certDigestValue!=null){
				Element certDigest=makeEasyElement("CertDigest",namespace);
				Element digestMethod=makeEasyElement("DigestMethod",SignedDoc.XMLNS_XMLDSIG);
				digestMethod.addAttribute(new Attribute("Algorithm",m_certDigestAlgorithm));
				certDigest.appendChild(digestMethod);
				Element digestValue=makeEasyElement("DigestValue",SignedDoc.XMLNS_XMLDSIG);
				digestValue.appendChild(Base64Util.encode(m_certDigestValue, m_certDigestValue.length));
				certDigest.appendChild(digestValue);
				cert.appendChild(certDigest);
			}
			if (m_sig.getKeyInfo()!=null&&m_certSerial!=null){
				Element issuerSerial=makeEasyElement("IssuerSerial",namespace);
				Element issuerName=makeEasyElement("X509IssuerName",SignedDoc.XMLNS_XMLDSIG);
				Principal principal=m_sig.getKeyInfo().getSignersCertificate().getIssuerDN();
//				X500Principal principal=m_sig.getKeyInfo().getSignersCertificate().getIssuerX500Principal();
				
				issuerName.appendChild(principal.getName());
				Element serialNumber=makeEasyElement("X509SerialNumber",SignedDoc.XMLNS_XMLDSIG);
				serialNumber.appendChild(m_certSerial.toString());
				issuerSerial.appendChild(issuerName);
				issuerSerial.appendChild(serialNumber);
				cert.appendChild(issuerSerial);
			}
			signingCertificate.appendChild(cert);
			signedSignatureProps.appendChild(signingCertificate);
		}
		
//		thisObject.appendChild(signedSignatureProps);
		
		if(m_address != null) {
			Element address=m_address.createElement(namespace);
			if (address!=null) {
				address.detach();
				signedSignatureProps.appendChild(address);
			}
		}
		
		if (m_policy!=null) {
			Element policy=m_policy.createElement(namespace);
			if (policy!=null) {
				policy.detach();
				signedSignatureProps.appendChild(policy);
			}
		}
		if(countClaimedRoles() > 0) {
			Element signerRole=makeEasyElement("SignerRole",namespace);
			Element claimedRoles=makeEasyElement("ClaimedRoles",namespace);
			signerRole.appendChild(claimedRoles);
			Builder builder=XOMBuilderFactory.getInstance(false);
			for(int i = 0; i < countClaimedRoles(); i++) {
				Element claimedRole=makeEasyElement("ClaimedRole",namespace);
				String claimedRoleStr=getClaimedRole(i);
				try {
					Document doc=builder.build(claimedRoleStr, "");
					claimedRole.appendChild(doc.getRootElement().copy());
				}catch(Exception e) {
					claimedRole.appendChild(claimedRoleStr);
				}				
				claimedRoles.appendChild(claimedRole);
			}
			signedSignatureProps.appendChild(signerRole);
		}
		thisObject.appendChild(signedSignatureProps);
		return thisObject;
	}

	public DataObjectFormat getDataObjectFormat() {
		return m_dataObjectFormat;
	}

	public void setDataObjectFormat(DataObjectFormat objectFormat) {
		m_dataObjectFormat = objectFormat;
	}
	
	
}
